%                                                             
%------------------------------------------------------------------------------------------------------------------------------------------------- 
%
%                                  Probabilisting Learning  on Manifolds (PLoM) with Partition 
%                                                 Version  24 June 2021
%
%                                               Copyright C. Soize, 24 June 2021
%                            
%                                             MSME UMR 8208, Université Gustave Eiffel
%
%                                                christian.soize@univ-eiffel.fr
%                            https://pagespro.univ-gustave-eiffel.fr/christian-soize
%                            https://scholar.google.fr/citations?user=XktTSHMAAAAJ\&hl=fr
%                            http://msme.u-pem.fr/equipe-mecanique/les-publications-hal/halbib/soize    
%
%                                          --- Software written with Matlab 2020b ---
%-----------------------------------------------------------------------------------------------------------------------------------------------
%
% The software "Probabilisting Learning on Manifolds (PLoM) with Partition" is a novel version of the PLoM algorithm 
% for which the first version of the algorithm was published in Ref. [1] and for which the mathematics foundations 
% can be found in Ref. [2].
% The present version of this PLoM software with partition is based on Ref.[3] and includes four novel capabilities:
%                - probabilistic learning on manifolds with partition that consists (i) in computing, before the learning, 
%                  an optimal partition in terms of independent random vectors (groups) using the algorithm presented in 
%                  Ref.[4] and (ii) in performing the probabilistic learning for each group of the identified partition.
%                - parallel computing.
%                - automatic identification of the smoothing parameter of the DMAP kernel as explained in Ref.[3].
%                - possibility to introduce constraints for preserving the normalization of the PCA coordinates during the 
%                  probabilistic learning process as explained in Ref.[3], based on Ref.[5].
% 
% Publications: [1] C. Soize, R. Ghanem, Data-driven probability concentration and sampling on manifold, 
%                   Journal of Computational Physics,  doi:10.1016/j.jcp.2016.05.044, 321, 242-258 (2016).
%                            
%               [2] C. Soize, R. Ghanem, Probabilistic learning on manifolds, Foundations of Data Science, 
%                   American  Institute of Mathematical Sciences (AIMS), doi: 10.3934/fods.2020013, 2(3), 279-307 (2020).
%                   Also in arXiv:2002.12653  [math.ST], 28 Feb 2020, https://arxiv.org/abs/2002.12653.                                
%
%               [3] C. Soize, R. Ghanem, Probabilistic learning on manifolds with partition, 
%                   in arXiv:2010.14324 [stat.ML], 21 Feb 2021, https://arxiv.org/abs/2102.10894 
%                   Also submitted in SIAM-ASA Journal on Uncertainty Quantification}, 2021.
%                            
%               [4] C. Soize, Optimal partition in terms of independent random vectors of any non-Gaussian vector 
%                   defined by a set of realizations, SIAM-ASA Journal on Uncertainty Quantification, 
%                   doi: 10.1137/16M1062223, 5(1), 176-211 (2017).    
%
%               [5] C. Soize, R. Ghanem, Physics-constrained non-Gaussian probabilistic learning on manifolds, International 
%                   Journal for Numerical Methods in Engineering, doi: 10.1002/nme.6202, 121 (1), 110-145 (2020).
%
%  This version allows for reproducing Application 1 of the paper:
%  [3] C. Soize, R. Ghanem, Probabilistic learning on manifolds with partition,
%      in arXiv:2010.14324 [stat.ML], 21 Feb 2021, https://arxiv.org/abs/2102.10894. 
%      Also submitted in SIAM-ASA Journal on Uncertainty Quantification, 2021".
%
%  The input data parameters entered for each STEP correpond to those for Application 1 for which the results are
%  in the directory "Results_AP1"
%
%-----------------------------------------------------------------------------------------------------------------------------------------------
%

  clear all                                     
  status=fclose('all'); 
  
  rng('default')
                         %--- SAVE 
  isave    = 0;          % isave     = 0 no save
                         %           = 1 save after STEP 1 Data generation for XX 
                         %           = 2 save after STEP 2 PCA of XX for constructing H 
                         %           = 3 save after STEP 3 learning before computing independent group for random vector H (step 3 is not mandatory)
                         %           = 4 save after STEP 4 computing independent groups for random vector H from MatReta_d(n_d,N_d)
                         %           = 5 save after STEP 5 learning after computing independent groups for random vector H                          
                         %--- RESTORE 
  irestore = 0;          % irestore  = 0 no restore
                         %           = 1 restore just after STEP 1 Data generation for XX 
                         %           = 2 retore  just after STEP 2 PCA of XX for constructing H
                         %           = 3 restore just after STEP 3 learning before computing independent group for random vector H (step 3 is notmandatory)
                         %           = 4 restore just after STEP 4 computing independent groups for random vector H from MatReta_d(n_d,N_d)
                         %           = 5 restore just after STEP 5 learning after computing independent groups for random vector H                         
  isavebid    = isave;                          
  irestorebid = irestore;  
     if irestore == 1; load matlabSTEP1_PLoM_Indep_Group.mat; end         
     if irestore == 2; load matlabSTEP2_PLoM_Indep_Group.mat; end      
     if irestore == 3; load matlabSTEP3_PLoM_Indep_Group.mat; end    
     if irestore == 4; load matlabSTEP4_PLoM_Indep_Group.mat; end    
     if irestore == 5; load matlabSTEP5_PLoM_Indep_Group.mat; end   
  isave       = isavebid;                       
  irestore    = irestorebid;                    
  clear isavebid irestorebid      
                                                %--- TYPE OF EXEC
  iexec_Step1 = 1;            % iexec_Step1    = 0 no Step 1 
                              %                = 1    Step 1 Data generation for XX 
  iexec_Step2 = 1;            % iexec_Step2    = 0 no Step 2 
                              %                = 1    Step 2 PCA of XX for constructing H
  iexec_Step3 = 1;            % iexec_Step3    = 0 no Step 3 
                              %                = 1    Step 3 learning before computing independent group for random vector H (step 3 is not mandatory)
  iexec_Step4 = 1;            % iexec_Step4    = 0 no Step 4 
                              %                = 1    Step 4 computing independent groups for random vector H from MatReta_d(n_d,N_d) 
  iexec_Step5 = 1;            % iexec_Step5    = 0 no Step 5 
                              %                = 1    Step 5 learning after computing independent group for random vector H  
  iexec_debugg = 0;           % = 0 no debugg sequence
                              % = 1    debugg sequence  

%============================================================================================================================================
%                                         Step1: Data generation for XX 
%============================================================================================================================================                   
    
if iexec_Step1 == 1                                                     % Step 1 data generation
   if irestore < 1                                                      % irestore >= 1 : skip Step 1 
      disp('STEP 1 -----------------------') 
      TimeStart1 = tic;       
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'STEP 1 ------------------------------------ \n ');
      fprintf(fidlisting,'      \n ');                                         %                                                                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting); 
      
                                                                 %--- DATA GENERATION FOR APPLICATION 1
      ind_application = 1;                                                           
      
      N_d = 1200;                                                %    number of realizations  
      
      [n_d,nuGen,MatRetaGen] = sub_gener_DATA_APPLICATION_1(N_d); 
      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'--- DATA PARAMETERS FOR APPLICATION %2i \n ', ind_application);   
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,' n_d       = %7i \n ',n_d);
      fprintf(fidlisting,' N_d       = %7i \n ',N_d);
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fclose(fidlisting); 
      
      ElapsedTime1 = toc(TimeStart1);      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'-------   ELAPSED TIME FOR STEP 1\n ');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Elapsed Time   =  %10.2f\n',ElapsedTime1);   
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);                      
      
      SAVERAND = rng;
      
      if isave == 1                                 %--- SAVE after STEP 1 ..............................
         close all                                  %
         save matlabSTEP1_PLoM_Indep_Group.mat -v7.3 
      end                                           %
   end                                              %--- End of ``if irestore < 1 ''
end                                                 %--- End of ``if iexec_Step1 == 1 ''  

%============================================================================================================================================
%                                 STEP 2 :  DATA for plot of XX_ar in Step3  and XX_arg in Step 5
%                                           DATA for PCA and construction of H as PCA of X = scaling of XX
%============================================================================================================================================
                                                     
if iexec_Step2 == 1                                  % Step 2
   if irestore < 2                                   % irestore >= 2 : skip Step 2  
      disp('STEP 2 -----------------------') 
      
      TimeStart2 = tic; 
      rng(SAVERAND);
      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'STEP 2 ------------------------------------ \n ');
      fprintf(fidlisting,'      \n ');                                         %                                                                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);  
                                                    %--- Data for PCA in Step 2
      ind_PCA    = 0;                               % = 0 : no PCA is performed and MatReta_d = MatRetaGen (case for Application 1)
                                                    % = 1 : PCA is performed on MatRx_d
      error_PCA  = 1e-06;                           %     : not used for Application 1
                                                    %
      ind_scale  = 0;                               % = 0 : no scaling and MatRx_d = MatRxx_d (case for Application 1)
                                                    % = 1 : scaling MatRxx_d yielding MatRx_d
                                                    %    
                                                    %--- Data for pdf and plot of XX_ar in Step 3 and XX_arg in Step5
                                                    %    if n_d is very high and if nbplotXX > 0 and/or nbplotXXClouds > 0, 
                                                    %    the array Matxx_d is kept in the RAM and can produce an out of memory.
      RplotXX             = [ ]';        
      nbplotXX            = size(RplotXX,1); 
      
      MatRplotXXClouds    = [ ];
      nbplotXXClouds      = size(MatRplotXXClouds,1);    
      
      fidlisting=fopen('listing.txt','a+');  
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'ind_PCA        = %6i \n ',ind_PCA); 
      fprintf(fidlisting,'                     \n ');  
      fprintf(fidlisting,'error_PCA      = %14.7e \n ',error_PCA); 
      fprintf(fidlisting,'                     \n '); 
      fprintf(fidlisting,'ind_scale      = %6i \n ',ind_scale); 
      fprintf(fidlisting,'                     \n '); 
      fprintf(fidlisting,'nbplotXX       = %6i \n ',nbplotXX); 
      fprintf(fidlisting,'                      \n '); 
      fprintf(fidlisting,'nbplotXXClouds = %6i \n ',nbplotXXClouds); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting); 
                                                    %--- checking data
                                                    
      if ind_PCA == 0 && (nbplotXX > 0 || nbplotXXClouds > 0)
         error('STOP: if nbplotXX > 0 or nbplotXXClouds > 0, ind_PCA must be equal to 1')
      end
      
      if ind_PCA == 0 && ind_scale ==1
         error('STOP: if ind_scale = 1, ind_PCA must be equal to 1')
      end
      
      if nbplotXX >= 1                              
         for iplotXX = 1:nbplotXX 
             ih = RplotXX(iplotXX);
             if ih >  n_d
                error('STOP: A component of XX, entered for plotting its additional realizations is out of range 1 to N_d')
             end
         end
      end

      if nbplotXXClouds > 1
         for iplotXXClouds = 1:nbplotXXClouds
             ih = MatRplotXXClouds(iplotXXClouds,1);
             jh = MatRplotXXClouds(iplotXXClouds,2);
             kh = MatRplotXXClouds(iplotXXClouds,3); 
             if ih > n_d || jh > n_d || kh > n_d
                 error('STOP: error in plot clouds data for XX')
             end
         end
      end
      
      %---------------------------------------------------------------------------------------------------------------------------------------                   
      %                   Scaling the initial dataset: construction of Ralpha, Ralpham1, Rbeta, MatRx_d, RalphaW, RbetaW 
      %---------------------------------------------------------------------------------------------------------------------------------------  
      
      if nbplotXX  == 0 &&  nbplotXXClouds == 0 
         MatRxx_d = [];
         MatRx_d = [];
         Rbeta = [];
         Ralpha = [];
         Ralpham1 = [];   
      end
      
      if nbplotXX  > 0 ||  nbplotXXClouds > 0  
         
         if ind_scale == 0
            MatRx_d  = MatRxx_d;                                                %    MatRx_d(n_d,N_d), MatRxx_d(n_d,N_d)
            Ralpha   = ones(n_d,1);
            Rbeta    = zeros(n_d,1);
         end
         if ind_scale == 1
            Rmax     = max(MatRxx_d,[],2);                                      %    MatRxx_d(n_d,N_d)
            Rmin     = min(MatRxx_d,[],2);
            MatRx_d  = zeros(n_d,N_d);                                          %    MatRx_d(n_d,N_d): scaling of MatRxx_d loaded in MatRx_d
            Rbeta    = zeros(n_d,1);                                            %    Rbeta(n_d,1);
            Ralpha   = zeros(n_d,1);                                            %    Ralpha(n_d,1);
            Ralpham1 = zeros(n_d,1);                                            %    Ralpham1(n_d,1);
            for i = 1:n_d
                if Rmax(i)-Rmin(i) ~= 0
                   Rbeta(i)     = Rmin(i);
                   Ralpha(i)    = Rmax(i)-Rmin(i);
                   Ralpham1(i)  = 1/Ralpha(i);
                   MatRx_d(i,:) = Ralpham1(i)*(MatRxx_d(i,:) - Rbeta(i)); 
                else
                   Ralpha(i)    = 1;
                   Ralpham1(i)  = 1; 
                   Rbeta(i)     = Rmin(i);
                   MatRx_d(i,:) = MatRxx_d(i,:) - Rbeta(i); 
                end
            end 
            clear Rmax Rmin
         end
      end                                              
      
      %--------------------------------------------------------------------------------------------------------------------------------------                   
      %                                PCA of MatRx_d(n_d,N_d) (scaled initial data)  
      %--------------------------------------------------------------------------------------------------------------------------------------  
      
                                       % MatRx_d(n_d,N_d) (scaled initial data)
                                       % MatReta_d(nu,N_d)
      if ind_PCA == 0                                 
         nu        = nuGen;
         nnull     = 0;
         MatReta_d = MatRetaGen;  
         
         fidlisting=fopen('listing.txt','a+');  
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'PCA is not performed, and MatReta_d = MatRetaGen'); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'Dimension nu of the random vector H = %4i \n ',nu); 
         fprintf(fidlisting,'      \n ');  
         fprintf(fidlisting,'      \n '); 
         fclose(fidlisting); 
      end
      
      if ind_PCA == 1  
                                                                         % nu = n_d - nnull, Rmu(nu,1), MatReta_d(nu,N_d), MatRVector(n_d,nu) 
         [nu,nnull,MatReta_d,Rmu,MatRVector] = sub_PCA_function(n_d,N_d,MatRx_d,error_PCA);  
         
         MatRdiagmu1s2 = diag(sqrt(Rmu));                                %    MatRdiagmu1s2(nu,nu)                                                                              
         RXmean        = mean(MatRx_d,2);                                %    RXmean(n_d,1)
         MatRphiXmu1s2 = MatRVector*MatRdiagmu1s2;                       %    MatRphiXmu1s2(n_d,nu) 
         clear MatRVector MatRdiagmu1s2
                                                                         %--- computing the L2 error and the second-order moments of X 
         MatRXmean = repmat(RXmean,1,N_d);                               %    MatRXmean(n_d,N_d)                                                                           
         MatRX_nu = MatRXmean + MatRphiXmu1s2*MatReta_d;                 %    MatRX_nu(n_d,N_d)
         error_nu = norm(MatRx_d-MatRX_nu,'fro')/norm(MatRx_d,'fro'); 

         clear MatRXmean MatRX_nu 
                                                                         %--- Print and plot
         fidlisting=fopen('listing.txt','a+');  
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n ');                     
         fprintf(fidlisting,'error_PCA = %14.7e \n ',error_PCA); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'Null-space dimension nnull of the covariance matrix of scaled X = %4i \n ',nnull); 
         fprintf(fidlisting,'Dimension nu of the random vector H = %4i \n ',nu); 
         fprintf(fidlisting,'      \n ');  
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'Rmu =          \n '); 
         fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',Rmu');
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'L2 error due to the PCA of scaled X: error_nu   = %14.7e \n ',error_nu); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fclose(fidlisting); 
      end
      
      ind_print_d = 0;
      if ind_print_d == 1                                                      %--- the print is bypassed if ind_print_d = 0 
         RHmean    = mean(MatReta_d,2);                                        %    RHmean(nu,1)
         MatRHcov  = cov(MatReta_d');                                          %    MatRHcov(nu,nu)  
         fidlisting=fopen('listing.txt','a+');  
         fprintf(fidlisting,'MEAN VECTOR AND COVARIANCE MATRICES OF  H  \n ');                        
         fprintf(fidlisting,'               \n ');
         fprintf(fidlisting,'               \n ');  
         fprintf(fidlisting,'RHmean =          \n '); 
         fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',RHmean');
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'MatRHcov =          \n '); 
         for k=1:nu       
             fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',MatRHcov(k,:));
             fprintf(fidlisting,'      \n ');
         end   
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n ');       
         fclose(fidlisting); 
         
         clear RHmean MatRHcov  
      end      
      
      ElapsedTime2 = toc(TimeStart2);      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'-------   ELAPSED TIME FOR STEP 2\n ');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Elapsed Time   =  %10.2f\n',ElapsedTime2);   
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);                        
      
      SAVERAND          = rng;
      SAVERANDfromStep2 = rng;
      
      if isave == 2                                 %--- SAVE after STEP 2 ..............................
         close all                                  %
         save  matlabSTEP2_PLoM_Indep_Group.mat -v7.3 
      end
   end                                              %--- End of ``if irestore < 2 ''
end                                                 %--- End of ``if iexec_Step2 == 1 ''  
      
%==========================================================================================================================================
%                    STEP 3 :  learning before computing independent group for random vector H (step 3 is not mandatory)
%==========================================================================================================================================
                                                     
if iexec_Step3 == 1                                  % Step 3
   if irestore < 3                                   % irestore >= 3 : skip Step 3  
      disp('STEP 3 -----------------------') 
      
      TimeStart3 = tic; 
      rng(SAVERAND);
      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'STEP 3 ------------------------------------ \n ');
      fprintf(fidlisting,'      \n ');                                         %                                                                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);  
      
      %-------------------------------------------------------------------------------------------------------------------------------------                           
      %                     Data input used for Step3 
      %-------------------------------------------------------------------------------------------------------------------------------------  
      
      epsilonDIFFmin      = 180;  % for mDMAP > 20, epsilonDIFF is searched in interval [epsilonDIFFmin , +infty[. The value of epsilonDIFFmin
                                  %                 can be estimated thanks to the graph of Rcomp (it is recommended to start with a 
                                  %                 sufficiently large value of  epsilonDIFFmin (for instance 2*nu ou 3*nu) and then 
                                  %                 ajusting epsilonDIFF using the graph of Rcomp.
      tdiffusion          = 0; 
      icorrectif          = 1;
      Facteur             = 20;
      f0_ref              = 4;            % reference value
      ind_f0              = 0;            % if ind_f0 = 0, then f0=f0_ref   and   if ind_f0 = 1, then f0=f0_ref/sh
      M0transient         = 30;
      nbMC                = 1000;
      NnbMC               = N_d*nbMC;
      ind_generator       = 1;  
     
      ind_constraints     = 0;    % = 0 : no constraints E{H] = 0 and E{H H'} = [I_nu]
                                  % = 1 : constraints E{H_j^2} = 1 for j =1,...,nu      (HIGHLY RECOMMENDED)
                                  % = 2 : constraints E{H] = 0 and E{H_j^2} = 1 for j =1,...,nu
                                  % = 3 : constraints E{H] = 0 and E{H H'} = [I_nu]                                
                                  %--- parameters for the convergence of the iteration algorithm if ind_constraints >= 1
          ind_coupling  = 0;      % = 0 : for ind_constraints = 2 or 3, no coupling in  matrix MatRGammaS_iter (HIGHLY RECOMMENDED)
                                  % = 1 : for ind_constraints = 2 or 3, coupling,  all the extra-diagonal blocs in matrix MatRGammaS_iter are kept
          epsc          = 1e-4;   %    relative tolerance for the iteration convergence in the constraints E{H} = 0 and E{H H'} = [I_nu] 
          iter_limit    = 700;    %    maximum number of iteration for computing the Lagrange multiplier
                                  %... Defining the relaxation parameter Ralpha_relax_step3(iter_limit,1)
          alpha_relax1   = 0.01;  %    value of alpha_relax for iter = 1  
          iter_relax2    = 20;    %    value of iter for which alpha_relax = alpha_relax2; if iter_relax2 = 1, then alpha_relax = alpha_relax2 for iter >=1   
          alpha_relax2   = 0.5;   %    value of alpha_relax for iter >= iter_relax2 
                                  %    NOTE 1: IF  iter_relax2 = 1 , THEN: Ralpha_relax_step3(iter) = alpha_relax2 for iter >=1
                                  %    NOTE 2: IF  iter_relax2 >= 2, THEN  
                                  %                        for iter >= 1 and for iter < iter_relax2, we have:
                                  %                            Ralpha_relax_step3(iter) = alpha_relax1 +(alpha_relax2 - alpha_relax1)*(iter-1)/(iter_relax2-1);
                                  %                        for iter >= iter_relax2, we have:
                                  %                            Ralpha_relax_step3(iter) = alpha_relax2
                                  %    NOTE 3: for decreasing the error err(iter), increase the value of iter_relax2
                                  %    NOTE 4: if divergence during ITO solver, decrease alpha_relax2 or increase iter_relax2
                                  %  
          minVarH  = 0.99;        %    minVarH          = minimum imposed on Var{H^2} with respect to 1 (for instance 0.99) for the convergence test
          maxVarH  = 1.01;        %    maxVarH          = maximum imposed on Var{H^2} with respect to 1 (for instance 1.01) for the convergence test 
                                  %
      irenormalization    = 0;  % = 0 : no re-normalization after learning (HIGHLY RECOMMENDED if ind_constraints >= 1)
                                % = 1 : re-normalization after learning in order that E{H_ar]=0 and E{H_ar H_ar'}=[I_nu] 
                                % = 2 : re-normalization after learning in order that E{H_ar]=0 
 
      RplotH              = [1 2 3 4 5 6 7]';         % for samples and for pdf of Har
      nbplotH             = size(RplotH,1); 
      MatRplotHClouds     = [1 2 3                    % clouds Har
                             2 3 4
                             3 4 5
                             4 5 6
                             5 6 7];
      nbplotHClouds       = size(MatRplotHClouds,1);  

      %-------------------------------------------------------------------------------------------------------------------------------------------                           
      %                     Checking data input and data plots
      %-------------------------------------------------------------------------------------------------------------------------------------------     
      if ind_constraints >= 1   
         if alpha_relax1 < 0 || alpha_relax1 > 1
            error('STOP1 : error in the value of alpha_relax1')
         end
         if alpha_relax2 < 0 || alpha_relax2 > 1
            error('STOP2 : error in the value of alpha_relax2')
         end
         if iter_relax2 >= 2 && iter_relax2 <= iter_limit
            if alpha_relax1 >  alpha_relax2
                error('STOP3 : error in the value of alpha_relax1 or alpha_relax2 or iter_relax2')
            end
         end
         if iter_relax2 > iter_limit
             error('STOP4 : error in the value of iter_relax2')
         end
                                            %--- Generation of Ralpha_relax_step3(iter_limit,1)
         Ralpha_relax_step3 = zeros(iter_limit,1);                                   
         if iter_relax2 == 1
            Ralpha_relax_step3(1:iter_limit) = alpha_relax2;
         end
         if iter_relax2 >= 2
            for iter = 1:iter_limit
                if iter >= 1 && iter < iter_relax2
                   Ralpha_relax_step3(iter) = alpha_relax1 +(alpha_relax2 - alpha_relax1)*(iter-1)/(iter_relax2-1);
                end
                if iter >= iter_relax2
                   Ralpha_relax_step3(iter) = alpha_relax2;
                end
            end
         end
      end        
                                         
      if nbplotH >= 1                              
         for iplotH = 1:nbplotH 
             ih = RplotH(iplotH);
             if ih >  nu
                error('STOP: A component of H, entered for plotting its additional realizations is out of range 1 to nu')
             end
         end
      end

      if nbplotHClouds > 1
         for iplotHClouds = 1:nbplotHClouds
             ih = MatRplotHClouds(iplotHClouds,1);
             jh = MatRplotHClouds(iplotHClouds,2);
             kh = MatRplotHClouds(iplotHClouds,3); 
             if ih > nu || jh > nu || kh > nu
                 error('STOP: error in plot clouds data for H')
             end
         end
      end  

      %---------------------------------------------------------------------------------------------------------------------------------
      %                                    Computing epsilonDIFF,mDP,mDMAP,nbmDMAP,Indg,s,shss,sh,Deltar,M0estim,ximean,f0 
      %---------------------------------------------------------------------------------------------------------------------------------  
      
      [epsilonDIFF,mDP,mDMAP,nbmDMAP,Indg,s,shss,sh,Deltar,M0estim,ximean,f0] = ...
                                  sub_diffusion_maps_manifold_for_parallel_parameters(nu,N_d,MatReta_d,ind_generator,icorrectif, ...
                                                                                      Facteur,f0_ref,ind_f0,epsilonDIFFmin); 
          
      %---------------------------------------------------------------------------------------------------------------------------------                           
      %                                    Print data input for learning 
      %---------------------------------------------------------------------------------------------------------------------------------  

      fidlisting=fopen('listing.txt','a+'); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');                     
      fprintf(fidlisting,'--- GENERAL DATA PARAMETERS FOR THE LEARNING STEP 3\n ');   
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,' n_d           = %7i \n ',n_d); 
      fprintf(fidlisting,' N_d           = %7i \n ',N_d);
      fprintf(fidlisting,' nu            = %7i \n ',nu);
      fprintf(fidlisting,' nnull         = %7i \n ',nnull);
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'epsilonDIFFmin = %14.7e \n ',epsilonDIFFmin); 
      fprintf(fidlisting,'epsilonDIFF    = %14.7e \n ',epsilonDIFF); 
      fprintf(fidlisting,'tdiffusion     = %8.4f \n ',tdiffusion);
      fprintf(fidlisting,'mDP            = %7i \n ',mDP); 
      fprintf(fidlisting,'mDMAP          = %7i \n ',mDMAP); 
      fprintf(fidlisting,'nbmDMAP        = %7i \n ',nbmDMAP); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'Indg =          \n '); 
      fprintf(fidlisting,' %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i \n ',Indg');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'icorrectif  = %2i \n ',icorrectif); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Facteur     = %4i \n ',Facteur);     
      fprintf(fidlisting,'f0_ref      = %8.4f \n ',f0_ref); 
      fprintf(fidlisting,'ind_f0      = %7i \n ',ind_f0);
      fprintf(fidlisting,'f0          = %8.4f \n ',f0); 
      fprintf(fidlisting,'Deltar      = %14.7e \n ',Deltar); 
      fprintf(fidlisting,'      \n ');
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'s           = %14.7e \n ',s); 
      fprintf(fidlisting,'shss        = %14.7e \n ',shss); 
      fprintf(fidlisting,'sh          = %14.7e \n ',sh); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'xi_mean = %14.7e \n ',ximean); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'M0estim = %7i \n ',M0estim); 
      fprintf(fidlisting,'M0transient = %7i \n ',M0transient); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'nbMC        = %7i \n ',nbMC); 
      fprintf(fidlisting,'NnbMC       = %7i \n ',NnbMC); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'ind_generator    = %1i \n ',ind_generator); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'irenormalization = %1i \n ',irenormalization); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'ind_constraints  = %1i \n ',ind_constraints);
      fprintf(fidlisting,'    ind_coupling = %7i \n ',ind_coupling);
      fprintf(fidlisting,'    epsc         = %14.7e \n ',epsc); 
      fprintf(fidlisting,'    iter_limit   = %7i \n ',iter_limit); 
      fprintf(fidlisting,'    alpha_relax1  = %8.4f \n ',alpha_relax1); 
      fprintf(fidlisting,'    iter_relax2   = %7i \n ',iter_relax2); 
      fprintf(fidlisting,'    alpha_relax2  = %8.4f \n ',alpha_relax2); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);  

      %--------------------------------------------------------------------------------------------------------------------------------------                           
      %                                    Generating NnbMC realizations 
      %--------------------------------------------------------------------------------------------------------------------------------------  
       
                                                                  %--- Generation of random Wienner germs for the Ito equation
      ArrayGauss              = randn(nu,N_d,nbMC);               %    ArrayGauss(nu,N_d,nbMC)
      ArrayWiennerM0transient = randn(nu,N_d,M0transient,nbMC);   %    ArrayWiennerM0transient(nu,N_d,M0transient,nbMC) 
      
      if ind_constraints == 0                                     %--- the constraints E{H] = 0 and E{H H'} = [I_nu] is not applied
                                                                  %    MatReta_ar(nu,NnbMC)                                             
         [MatReta_ar]  = sub_diffusion_maps_manifold_for_parallel(nu,N_d,MatReta_d,epsilonDIFF,tdiffusion,mDP,mDMAP,nbmDMAP,Indg,f0, ...
                                        M0transient,nbMC,NnbMC,ind_generator,shss,sh,Deltar,ArrayWiennerM0transient,ArrayGauss,irenormalization); 
      end
      
      if ind_constraints >= 1                                     %--- constraints are applied
                                                                  %    MatReta_ar(nu,NnbMC)                                            
         [MatReta_ar]  = sub_diffusion_maps_manifold_for_parallel_c(nu,N_d,MatReta_d,epsilonDIFF,tdiffusion,mDP,mDMAP,nbmDMAP,Indg,f0, ...
                                        M0transient,nbMC,NnbMC,ind_generator,shss,sh,Deltar,ArrayWiennerM0transient,ArrayGauss, ...
                                        irenormalization,ind_constraints,ind_coupling,epsc,iter_limit,Ralpha_relax_step3,minVarH,maxVarH); 
      end
      
      clear  ArrayGauss ArrayWiennerM0transient 
                                                         %--- Estimation of the measure concentration computing the L2 norm 
                                                         %    d2mopt_ar = E{|| [H_ar] - [eta_d] ||^2} / || [eta_d] ||^2
      d2mopt_ar  = 0; 
      ArrayH_ar  = reshape(MatReta_ar,nu,N_d,nbMC);      %    ArrayH_ar(nu,N_d,nbMC), MatReta_ar(nu,NnbMC)
      for ell = 1:nbMC
          d2mopt_ar = d2mopt_ar + (norm(ArrayH_ar(:,:,ell) - MatReta_d,'fro'))^2; %    MatReta_d(nu,N_d)
      end
      clear ArrayH_ar
      deno2     = (norm(MatReta_d,'fro'))^2;
      d2mopt_ar = d2mopt_ar/(nbMC*deno2);
      
      disp('------  d2mopt_ar = '); disp(d2mopt_ar)
      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'-------   L2 norm related to the concentration of the measure \n ');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'          Estimation of E{[H_ar]}: d^2(m_opt)_ar  =  %14.7e \n',d2mopt_ar); 
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);    
      
      %-----------------------------------------------------------------------------------------------------------------------------------------                           
      %         plot the addditional realizations of random vectors MatReta_ar(nu,NnbMC) and MatRxx_ar(n_d,NnbMC) generated by the PLoM  
      %----------------------------------------------------------------------------------------------------------------------------------------- 

      sub_plot_diffusion_maps_manifold_H_ar(n_d,N_d,NnbMC,MatReta_d,MatReta_ar,nbplotH,RplotH,nbplotHClouds,MatRplotHClouds);  
      
      if nbplotXX  > 0 ||  nbplotXXClouds > 0                      %--- construction of MatRxx_ar(n_d,NnbMC)
         MatRXmean = repmat(RXmean,1,NnbMC);                       %    MatRXmean(n_d,NnbMC) 
         MatRx_ar  = MatRXmean + MatRphiXmu1s2*MatReta_ar;         %    MatRXmean(n_d,NnbMC), MatRphiXmu1s2(n_d,nu),MatReta_arg(nu,NnbMC) 
                                                                   %--- Back scaling:  MatRxx_ar(n_d,NnbMC)
         MatRxx_ar = zeros(n_d,NnbMC);
         for k =1:n_d                                           
             MatRxx_ar(k,:) = Ralpha(k)*MatRx_ar(k,:) + Rbeta(k);   
         end  
         clear MatRx_ar
         sub_plot_diffusion_maps_manifold_XX_ar(n_d,N_d,NnbMC,MatRxx_d,MatRxx_ar,nbplotXX,RplotXX,nbplotXXClouds,MatRplotXXClouds);
      end
      
      ElapsedTime3 = toc(TimeStart3);      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'-------   ELAPSED TIME FOR STEP 3\n ');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Elapsed Time   =  %10.2f\n',ElapsedTime3);  
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);                        
      
      SAVERAND = rng;
      
      if isave == 3                                 %--- SAVE after STEP 3 ..............................
         close all                                  %
         save  matlabSTEP3_PLoM_Indep_Group.mat -v7.3     
      end                                           %
   end                                              %--- End of ``if irestore < 3 ''
end                                                 %--- End of ``if iexec_Step3 == 1 ''  
   
  
%===============================================================================================================================================
%                      Step 4: computing independent groups for random vector H from MatReta_d(n_d,N_d)
%===============================================================================================================================================

if iexec_Step4 == 1                                  % Step 4
   if irestore < 4                                   % irestore >= 4 : skip Step 4  
      disp('STEP 4 -----------------------') 
      
      TimeStart4 = tic; 
    
      rng(SAVERANDfromStep2);                        % the random generator is initialized with its value at the end of Step2 because 
                                                     % we must have the same state whether Step3 is used or not for running Step4
      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'STEP 4 ------------------------------------ \n ');
      fprintf(fidlisting,'      \n ');                                         %                                                                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);  
      
      numfig      = 0;
      iexec_plot2 = 1;     % = 0 no plot of joint pdf of  H_j with H_j'
                           % = 1    plot of joint pdf of  H_j with H_j'
                           
      RINDEPref =(0.001:0.001:0.020)';         
      nref = size(RINDEPref,1); 
      NKL  = nu;                                                                                          
      nr   = N_d;    
      RHplot           = [1 2 3 4 5 6 7];
      MatRcoupleRHplot = [1 2; 2 3; 3 4; 4 5; 5 6; 6 7];       
     
            
      MatRHexp = MatReta_d;           % MatRHexp(NKL,nr), MatReta_d(nu,N_d)
   
                                      %---------------------------------------------------------------------------------------------------
                                      %  Construction of independent subsets (groups) of the set of all the components of random vector H
                                      %---------------------------------------------------------------------------------------------------
                                      %
                                      % Decomposition of H = (H_1,...,H_r,...,H_NKL) in ngroup subsets (groups) H^1,...,H^j,...,H^ngroup
                                      % such that H =(H^1,...,H^j,...,H^ngroup) with j = 1,...,ngroup
                                      % and where H^j =(H_rj1,...,Hrjmj) 
                                      % with n1+ ... + nngroup = NKL
                                                
                                      %--- CONSTRUCTING A DECOMPOSITION IN GROUPS 
                                                
              %   ngroup = number of groups that are constructed
              %   Igroup = vector (ngroup,1) such that Igroup(j): number mj of the components of  Y^j = (H_jr1,... ,H_jrmj)
              %   mmax = max_j Igroup(j)
              %   MatIgroup = matrix(ngroup,mmax) such that MatIgroup1(j,r) = rj : indice rj de H dans le groupe j tel que Y^j_r = H_jrj
              %   nedge: number of edges in the graph
              %   MatPrint  = matrix(nedge,5) such that MatPrint(edge,:) = [edge r1 r2 INDEPr1r2 INDEPGaussRef]              
   
              %   INDEPGaussRef: Numerical criterion for testing the independence of two normalized Gaussian random variables HGauss_r1 and 
              %                  HGauss_r2 by using the MUTUAL INFORMATION criterion. The random generator for randn is reinitialized 
              %                  for avoiding some variations on the statistical fluctuations in the computation of INDEPGaussRef with respect
              %                  to the different cases analyzed with iexec_TEST. Note that the independent realizations used for constructing 
              %                  INDEPGaussRef can be dependent of all the other random 
              %                  quantities constructed before this one.

                                      
      disp('--- beginning the construction of the Gauss reference for independent group')
      MatRHexpGauss = randn(NKL,nr);   % (NKL,nr) matrix of the nr independent realizations of HGauss = (HGauss_1,...,HGauss_NKL)

                                       %--- CONSTRUCTING THE OPTIMAL VALUE INDEPopt such that the rate tau(INDEPref) is maximum

     [INDEPGaussmax,meanINDEPGauss,stdINDEPGauss,gust,meanMaxINDEPGauss,stdMaxINDEPGauss,RINDEPGauss,numfig] = ...
                                  sub_constructing_GAUSS_reference(NKL,nr,MatRHexpGauss,numfig); 

      disp('--- end of the construction of the Gauss reference for independent group') 

      disp('--- beginning the optimization loop for independent group')    

      Rtau = zeros(nref,1);
      parfor iref = 1:nref
          INDEPref = RINDEPref(iref); 

          [ngroup,Igroup,MatIgroup] = sub_find_groups1(NKL,nr,MatRHexp,INDEPref);  %   Construction of the groups corresponding to the value INDEPref
                                                                                   %   Rate of independence of the constructed partition     
          [tau] = sub_checking_independence_constructed_groups1(NKL,nr,ngroup,Igroup,MatIgroup,MatRHexp,MatRHexpGauss);  
          Rtau(iref) = tau;
      end

      disp('--- end of the optimization loop for independent group')
                                                                                   % Print and plot
      h = figure; 
      plot(RINDEPref,Rtau,'-ob')
           title({['Graph of the rate \tau (i_{ref}) of mutual information'] ; ['for the partition obtained with the level i_{ref}']},'FontSize',12,'Interpreter','tex','FontWeight','normal')                                        
           xlabel(['i_{ref}'],'FontSize',12,'Interpreter','tex')                                                                 
           ylabel(['\tau (i_{ref})'],'FontSize',12,'Interpreter','tex')   
           saveas(h,'figure_STEP4_GROUP_tau.fig') 
      close(h);
      
      disp('--- beginning the construction of the independent groups')
      
      [tauopt,irefopt] = max(Rtau);                                               % Calculation of INDEPopt 
      INDEPopt = RINDEPref(irefopt);
                                       %--- CONSTRUCTING THE GROUPS CORRESPONDING TO THE OPTIMAL VALUE INDEPopt of INDEPref

      [ngroup,Igroup,MatIgroup,nedge,MatPrintEdge,nindep,MatPrintIndep,npair,MatPrintPair,RplotPair] =  sub_find_groups2(NKL,nr,MatRHexp,INDEPopt); 
 
      disp('--- end of the construction of the independent groups')
                                       %--- PRINT AND PLOT AND CHECKING

      Indic = 0;                       % if Indic = 0 at the end of the loop, then Rtau(iref) < or = tauopt for all iref
      for iref=1:nref
          if Rtau(iref) > tauopt
             Indic = 1;
          end
      end
      if Indic == 0 && tauopt < 0     % then H_1,...,H_n are mutually independent and Gaussian
         ngroup = NKL; 
         Igroup = ones(ngroup,1);
         MatIgroup = (1:1:ngroup)';
         [numfig] = sub_print_plot_groups1(INDEPopt,ngroup,Igroup,MatIgroup,npair,RplotPair,numfig); 

      else                            % then H_1,...,H_n are mutually dependent and not Gaussian       

         [numfig] = sub_print_plot_groups2(INDEPopt,ngroup,Igroup,MatIgroup,nedge,MatPrintEdge,nindep,MatPrintIndep,npair, ...
                                           MatPrintPair,RplotPair,numfig); 
      end

      [INDEPGausscheck,INDEPcheck,tauopt] = sub_checking_independence_constructed_groups2(NKL,nr,ngroup,Igroup,MatIgroup, ...
                                                                                          MatRHexp,MatRHexpGauss);  


      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'--------------- SUMMARIZING THE OPTIMAL PARTITION CONSTRUCTED IN INDEPENDENT GROUPS');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Optimal value INDEPopt of INDEPref used for constructing the optimal partition = %8.5i \n ',INDEPopt);
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Optimal value tauopt of tau corresponding to the construction of the optimal partition = %8.5i \n ',tauopt); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'INDEPGaussmax = %8.6f \n ',INDEPGaussmax); 
      fprintf(fidlisting,'meanINDEPGauss = %8.6f \n ',meanINDEPGauss);
      fprintf(fidlisting,'stdINDEPGauss = %8.6f \n ',stdINDEPGauss);
      fprintf(fidlisting,'gust = %8.6f \n ',gust);
      fprintf(fidlisting,'meanMaxINDEPGauss = meanINDEPGauss + gust*stdINDEPGauss = %8.6f \n ',meanMaxINDEPGauss); 
      fprintf(fidlisting,'stdMaxINDEPGauss = %8.6f \n ',stdMaxINDEPGauss);
      fprintf(fidlisting,'number of groups  = %7i \n ',ngroup);  
      fprintf(fidlisting,'maximum number of nodes in the largest group  = %7i \n ',max(Igroup));  
      fprintf(fidlisting,'mutual information for the identified decomposition INDEPcheck  = %8.6f \n ',INDEPcheck);  
      fprintf(fidlisting,'mutual information for a normalized Gaussian vector having the identified decomposition INDEPGausscheck  = %8.6f \n ',INDEPGausscheck);
      fprintf(fidlisting,'the groups are independent if INDEPcheck < = INDEPGausscheck');    
      fprintf(fidlisting,'      \n ');  
      fclose(fidlisting); 
                                              %-------------------------------------------------------------------------------------------------
                                              %  Plot densite de proba des H_j et des couples H_j- H_j'
                                              %-------------------------------------------------------------------------------------------------
      if iexec_plot2 == 1  

         ncoupleRHplot = size(MatRcoupleRHplot,1);
         nRHplot = size(RHplot,2);
         npoint = 300;
                                              %--- estimation et plot des densites de probabilite des composantes H_j
         for jj = 1:nRHplot
              j = RHplot(jj);
              h = figure;                                                                                                                      
              [Rpdfj,Retaj] = ksdensity(MatRHexp(j,:),'npoints',npoint);
              Rgaussj = normpdf(Retaj,0,1);   
              plot(Retaj,Rpdfj,'-k',Retaj,Rgaussj,'--k')
              title({['Graph of the non-Gaussian pdf (solid black) and its'] ; ['Gaussian approximation (dashed black) for $H_{',num2str(j),'}$']},'FontSize',16,'Interpreter','latex','FontWeight','normal')                                        
              xlabel(['$\eta_{',num2str(j),'}$'],'FontSize',16,'Interpreter','latex')                                                                
              ylabel(['$p_{H_{',num2str(j),'}}(\eta_{',num2str(j),'})$'],'FontSize',16,'Interpreter','latex')                 
              saveas(h,['figure_STEP4_GROUP_pdf_H',num2str(j),'.fig'])
              close(h);
         end
         clear Rpdfj Retaj Rgaussj
                                                   %--- estimation et plot des densites de probabilite conjointes des composantes  H_j et H_j'
                                                   %

         npoint = 300;
         npointp1 = npoint+1;     
         RHexpmin = zeros(NKL,1);                  % lower and upper bounds of the support of each marginal pdf des H_j estimated with ksdensity
         RHexpmax = zeros(NKL,1);
         for j =1:NKL
             Rbid = MatRHexp(j,:)';                      % Rbid(nr,1)
             [Rpdf,R]=ksdensity(Rbid, 'npoints',npoint); % ksdensity is only used for computing the bounds and the point sampling of each axis
             minR = min(R);
             maxR = max(R);
             s = min(maxR,-minR);
             RHexpmin(j) = -1.5*s;                       
             RHexpmax(j) = 1.5*s; 
         end 
         clear Rbid R
        
         for icouple=1:ncoupleRHplot 
             j1 = MatRcoupleRHplot(icouple,1);
             j2 = MatRcoupleRHplot(icouple,2);     
             D1 =(RHexpmax(j1)-RHexpmin(j1))/npoint;
             D2 =(RHexpmax(j2)-RHexpmin(j2))/npoint;
             [MatR1,MatR2] = meshgrid(RHexpmin(j1):D1:RHexpmax(j1),RHexpmin(j2):D2:RHexpmax(j2)); % MatR1(npointp1,npointp1),MatR2(npointp1,npointp1)
           
             MatR3H      = zeros(npointp1,npointp1);
             for mm =1:npointp1       
                 MatRR =[MatR1(mm,:)                                       % MatR1(npointp1,npointp1),MatR2(npointp1,npointp1),MatRHexp(NKL,nr)
                         MatR2(mm,:)];                                     % MatRR(2,npointp1)

                 MatRRHData = [ MatRHexp(j1,:)                             % MatRRHData(2,nr)
                                MatRHexp(j2,:)];                
                [RpdfH] = sub_ksdensity_mult(NKL,nr,2,npointp1,MatRRHData,MatRR);  % RpdfH(npointp1,1) 
                 MatR3H(mm,:) = RpdfH';
             end
            
             h = figure;           
             colormap(bone)
             axis('square')
             contourf(MatR1,MatR2,MatR3H)                   
                      title(['pdf H_{',num2str(j1),'} and H_{',num2str(j2),'}'],'FontSize',16,'Interpreter','tex','FontWeight','normal') 
                      xlabel(['\eta_{',num2str(j1),'}'],'FontSize',16,'Interpreter','tex')                                                                
                      ylabel(['\eta_{',num2str(j2),'}'],'FontSize',16,'Interpreter','tex')
                      colorbar;                                                     
             saveas(h,['figure_STEP4_GROUP_H',num2str(j1),'_H',num2str(j2),'.fig']);  
             close(h);
         end
         clear MatR1 MatR2 MatR3HGauss MatR3H RpdfHGauss MatRRHData RpdfH  bid gust h
         clear D1 D2 j1 j2 nrGauss npoint npointp1 MatRRHDataGauss MatRR MatR3H  
         clear INDEPGausscheck INDEPGaussmax INDEPcheck INDEPopt INDEPref MatPrintEdge MatPrintIndep  MatPrintPair MatRHexp  MatRHexpGauss         
         clear RHexpmax  RHexpmin RINDEPGauss RINDEPref Rpdf  RplotPair  Rtau icouple ih iref irefopt NKL nindep
         clear j jh jj k kh m1 maxR meanINDEPGauss meanMaxINDEPGauss minR mm nedge nnull npair nr nref s status stdINDEPGauss stdMaxINDEPGauss tau tauopt                      
      end                                %--- end iexec_plot2
      
   
      ElapsedTime4 = toc(TimeStart4);      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'-------   ELAPSED TIME FOR STEP 4\n ');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Elapsed Time   =  %10.2f\n',ElapsedTime4);   
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);                        
           
      SAVERAND = rng;
      
      if isave == 4                                 %--- SAVE after STEP 4 ..............................
         close all                                  %
         save  matlabSTEP4_PLoM_Indep_Group.mat -v7.3    
      end                                           %
   end                                              %--- End of ``if irestore < 4 ''
end                                                 %--- End of ``if iexec_Step4 == 1 ''  

%===============================================================================================================================================
%                      Step 5: learning after computing independent groups for random vector H  
%===============================================================================================================================================

if iexec_Step5 == 1                                  % Step 5
   if irestore < 5                                   % irestore >= 5 : skip Step 5  
      disp('STEP 5 -----------------------') 
      
      TimeStart5 = tic; 
      rng(SAVERAND);
      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'STEP 5 ------------------------------------ \n ');
      fprintf(fidlisting,'      \n ');                                         %                                                                                                                
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);  
      
      %-------------------------------------------------------------------------------------------------------------------------------------------                           
      %                     Data input used for Step5 
      %------------------------------------------------------------------------------------------------------------------------------------------- 
                                          
      epsilonDIFFmin        = 180;        % if indDMAPS = 1 and for mDMAP > 20, epsilonDIFF is searched in interval [epsilonDIFFmin , +infty[. 
                                          %                 The value of epsilonDIFFmin can be estimated thanks to the graph of Rcomp 
                                          %                 (it is recommended to start with a sufficiently large value of epsilonDIFFmin 
                                          %                 (for instance 2*nu ou 3*nu) and then ajusting epsilonDIFF using the graph of Rcomp.
      
      epsilonDIFFmin_groups = 180;        % if indDMAPS = 2 and for mDMAPj > 20, epsilonDIFFj is searched in interval [epsilonDIFFmin_groups , +infty[.
                                          %                 The value of epsilonDIFFmin_groups can be estimated thanks to the graph of 
                                          %                 Rcomp_groupj (it is recommended to  start with a sufficiently large value of 
                                          %                 epsilonDIFFmin_groups (for instance 2*nu) and then ajusting  epsilonDIFFj 
                                          %                 using the graph of Rcomp_groupj.
      tdiffusion            = 0;  
      icorrectif            = 1;
      Facteur               = 20;
      f0_ref                = 4;    
      ind_f0                = 0;          % if ind_f0 = 0, then f0j=f0_ref   and   if ind_f0 = 1, then f0j=f0_ref/shj for each group j    
      M0transient           = 30;
      nbMC                  = 1000;
      NnbMC                 = N_d*nbMC;
      ind_generator         = 1;  
      
      ind_bandwidth_group = 2;    % = 1 : use the bandwidth of Step 3 related to nu for each group j (NOT RECOMMENDED)
                                  %   2 : use the bandwidth related to mj of each group j (HIGHLY RECOMMENDED)
      ind_DMAPS           = 2;    %     : not use if ind_generator = 0 because [g] = [I_Nd]
                                  % = 1 : use of mDMAP and MatRg(nu,mDMAP) constructed in STEP 3 (not mandatory, MatRg(nu,mDMAP) re-constructed in STEP5)
                                  % = 2 : use of mDMAPj and MatRgj(mj,mDMAPj) constructed for each group j                            
      ind_constraints     = 1;    % = 0 : no constraints E{H] = 0 and E{H H'} = [I_nu]
                                  % = 1 : constraints E{H_j^2} = 1 for j =1,...,nu   
                                  % = 2 : constraints E{H] = 0 and E{H_j^2} = 1 for j =1,...,nu
                                  % = 3 : constraints E{H] = 0 and E{H H'} = [I_nu]                                  
                                  %--- parameters for the convergence of the iteration algorithm if ind_constraints >= 1
          ind_coupling  = 0;      % = 0 : for ind_constraints = 2 or 3, no coupling in  matrix MatRGammaS_iter (HIGHLY RECOMMENDED)
                                  % = 1 : for ind_constraints = 2 or 3, coupling,  all the extra-diagonal blocs in matrix MatRGammaS_iter are kept
          epsc          = 1e-4;   %    relative tolerance for the iteration convergence in the constraints E{H} = 0 and E{H H'} = [I_nu] 
          iter_limit    = 700;    %    maximum number of iteration for computing the Lagrange multiplier
                                  %... Defining the relaxation parameter Ralpha_relax_step5(iter_limit,1)
          alpha_relax1   = 0.01;  %    value of alpha_relax for iter = 1  
          iter_relax2    = 20;    %    value of iter for which alpha_relax = alpha_relax2; if iter_relax2 = 1, then alpha_relax = alpha_relax2 for iter >=1   
          alpha_relax2   = 0.5;   %    value of alpha_relax for iter >= iter_relax2 
                                  %    NOTE 1: IF  iter_relax2 = 1 , THEN: Ralpha_relax_step5(iter) = alpha_relax2 for iter >=1
                                  %    NOTE 2: IF  iter_relax2 >= 2, THEN  
                                  %                        for iter >= 1 and for iter < iter_relax2, we have:
                                  %                            Ralpha_relax_step5(iter) = alpha_relax1 +(alpha_relax2 - alpha_relax1)*(iter-1)/(iter_relax2-1);
                                  %                        for iter >= iter_relax2, we have:
                                  %                            Ralpha_relax_step5(iter) = alpha_relax2
                                  %    NOTE 3: for decreasing the error err(iter), increase the value of iter_relax2
                                  %    NOTE 4: if divergence during ITO solver, decrease alpha_relax2 or increase iter_relax2
          minVarH  = 0.99;        %    minVarH          = minimum imposed on Var{H^2} with respect to 1 (for instance 0.99) for the convergence test
          maxVarH  = 1.01;        %    maxVarH          = maximum imposed on Var{H^2} with respect to 1 (for instance 1.01) for the convergence test                          
                                  %     
      irenormalization    = 0;    % = 0 : no re-normalization after learning (HIGHLY RECOMMENDED if ind_constraints >= 1)
                                  % = 1 : re-normalization after learning in order that E{H_ar]=0 and E{H_ar H_ar'}=[I_nu] (recommended if ind_constraints = 0)
                                  % = 2 : re-normalization after learning in order that E{H_ar]=0                                               
      
      RplotHgSamples      = [1 30 60]';                                  % samples for Hg 
      nbplotHgSamples     = size(RplotHgSamples,1); 
        
      RplotHg             = [1 2 3 4 5 6 7]';                              % pdf for Hg
      nbplotHg            = size(RplotHg,1); 
 
      MatRplotHgClouds    = [1 2 3                                      % clouds for Harg 
                             2 3 4
                             3 4 5
                             4 5 6
                             5 6 7];        
      nbplotHgClouds      = size(MatRplotHgClouds,1);  
                 
      %-------------------------------------------------------------------------------------------------------------------------------------------                           
      %                     Checking data input and data plots
      %-------------------------------------------------------------------------------------------------------------------------------------------     
      
      if ind_DMAPS ~= 1 && ind_DMAPS ~= 2
         error('STOP1 : data error on ind_DMAPS')
      end
      if ind_bandwidth_group ~= 1 && ind_bandwidth_group ~= 2
         error('STOP2 : data error on ind_bandwidth_group')
      end
      
      if ind_constraints >= 1   
         if alpha_relax1 < 0 || alpha_relax1 > 1
            error('STOP3 : data error on alpha_relax1')
         end
         if alpha_relax2 < 0 || alpha_relax2 > 1
            error('STOP4 : data error on alpha_relax2')
         end
         if iter_relax2 >= 2 && iter_relax2 <= iter_limit
            if alpha_relax1 >  alpha_relax2
                error('STOP5 : data error on alpha_relax1 or alpha_relax2 or iter_relax2')
            end
         end
         if iter_relax2 > iter_limit
             error('STOP6 : data error on iter_relax2')
         end
                                                             %--- Generation of Ralpha_relax_step5(iter_limit,1)
         Ralpha_relax_step5 = zeros(iter_limit,1);                                   
         if iter_relax2 == 1
            Ralpha_relax_step5(1:iter_limit) = alpha_relax2;
         end
         if iter_relax2 >= 2
            for iter = 1:iter_limit
                if iter >= 1 && iter < iter_relax2
                   Ralpha_relax_step5(iter) = alpha_relax1 +(alpha_relax2 - alpha_relax1)*(iter-1)/(iter_relax2-1);
                end
                if iter >= iter_relax2
                   Ralpha_relax_step5(iter) = alpha_relax2;
                end
            end
         end
      end
      
      if nbplotHgSamples >= 1                              
         for iplotH = 1:nbplotHgSamples 
             ih = RplotHgSamples(iplotH);
             if ih >  nu
                error('STOP: For samples, a component of Hg, entered for plotting its additional realizations is out of range 1 to nu')
             end
         end
      end
      
      if nbplotHg >= 1                              
         for iplotH = 1:nbplotHg 
             ih = RplotHg(iplotH);
             if ih >  nu
                error('STOP: A component of Hg, entered for plotting its additional realizations is out of range 1 to nu')
             end
         end
      end

      if nbplotHgClouds > 1
         for iplotHClouds = 1:nbplotHgClouds
             ih = MatRplotHgClouds(iplotHClouds,1);
             jh = MatRplotHgClouds(iplotHClouds,2);
             kh = MatRplotHgClouds(iplotHClouds,3); 
             if ih > nu || jh > nu || kh > nu
                 error('STOP: error in plot clouds data for Hg')
             end
         end
      end

      %-------------------------------------------------------------------------------------------------------------------------------------------                           
      %                                    Print data input for learning 
      %-------------------------------------------------------------------------------------------------------------------------------------------  

      fidlisting=fopen('listing.txt','a+'); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');                     
      fprintf(fidlisting,'--- GENERAL DATA PARAMETERS FOR THE LEARNING STEP 5\n ');   
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,' n_d         = %7i \n ',n_d); 
      fprintf(fidlisting,' N_d         = %7i \n ',N_d);
      fprintf(fidlisting,' nu          = %7i \n ',nu);
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'epsilonDIFFmin        = %14.7e \n ',epsilonDIFFmin); 
      fprintf(fidlisting,'epsilonDIFFmin_groups = %14.7e \n ',epsilonDIFFmin_groups); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'tdiffusion  = %8.4f \n ',tdiffusion);
      fprintf(fidlisting,'icorrectif  = %2i \n ',icorrectif); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Facteur     = %4i \n ',Facteur);  
      fprintf(fidlisting,'f0_ref      = %8.6f \n ',f0_ref);  
      fprintf(fidlisting,'ind_f0      = %7i \n ',ind_f0); 
      fprintf(fidlisting,'      \n ');      
      fprintf(fidlisting,'M0transient = %7i \n ',M0transient); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'nbMC        = %7i \n ',nbMC); 
      fprintf(fidlisting,'NnbMC       = %7i \n ',NnbMC); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'ind_generator    = %1i \n ',ind_generator); 
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'irenormalization = %1i \n ',irenormalization); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'ind_DMAPS           = %1i \n ',ind_DMAPS); 
      fprintf(fidlisting,'ind_bandwidth_group = %1i \n ',ind_bandwidth_group); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'ind_constraints  = %1i \n ',ind_constraints); 
      fprintf(fidlisting,'    ind_coupling = %7i \n ',ind_coupling);
      fprintf(fidlisting,'    epsc         = %14.7e \n ',epsc); 
      fprintf(fidlisting,'    iter_limit   = %7i \n ',iter_limit); 
      fprintf(fidlisting,'    alpha_relax1  = %8.4f \n ',alpha_relax1); 
      fprintf(fidlisting,'    iter_relax2   = %7i \n ',iter_relax2); 
      fprintf(fidlisting,'    alpha_relax2  = %8.4f \n ',alpha_relax2);   
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'      \n ');  
      fprintf(fidlisting,'      \n ');   
      fclose(fidlisting);  
      
      %---------------------------------------------------------------------------------------------------------------------------------------------------
      %                                     If ind_DMAPS = 1, computing the DMAPS basis as in  STEP 3
      %                                     Since STEP 3 is not mandatory,  nbmDMAP and MatRg(N_d,nbmDMAP) are re-computed in STEP5
      %---------------------------------------------------------------------------------------------------------------------------------------------------
 
      if ind_DMAPS == 1    %----- Computing epsilonDIFF,mDP,mDMAP,nbmDMAP,Indg,MatRg(N_d,nbmDMAP),Rlambda(mDP)  
         [epsilonDIFF,mDP,mDMAP,nbmDMAP,Indg,MatRg,Rlambda] = sub_diffusion_maps_parameters_step5(nu,N_d,MatReta_d,ind_generator,tdiffusion,epsilonDIFFmin); 
      end
      if ind_DMAPS == 2    %----- Not computing epsilonDIFF,mDP,mDMAP,nbmDMAP,Indg,MatRg  
         epsilonDIFF = [];
         mDP         = [];
         mDMAP       = [];
         nbmDMAP     = [];
         Indg        = [];
         MatRg       = [];
         Rlambda     = [];
      end
      
      %---------------------------------------------------------------------------------------------------------------------------------------------------
      %                                     Generating NnbMC realizations  on the independent groups
      %---------------------------------------------------------------------------------------------------------------------------------------------------
      
                                  %--- optimal partition in ngroup independent random vector Y^1,...,Y^ngroup of random vector H of dimension nu
                                  %    ngroup = number of groups Y^1,...,Y^ngroup
                                  %    Igroup = vector (ngroup,1) such that Igroup(j): number mj of the components of  Y^j = (H_jr1,... ,H_jrmj)
                                  %    MatIgroup = matrix(ngroup,mmax) such that MatIgroup1(j,r) = rj : indice rj de H dans le groupe j tel que Y^j_r = H_jrj
       
                                                                %--- construction of cellMatReta_d{j} = MatReta_dj(mj,NnbMC)
       cellMatReta_d        = cell(ngroup,1);                   %    cellMatReta_d{j} =  MatReta_dj(mj,NnbMC),  cellMatReta_d{ngroup}                          
       for j = 1:ngroup                                         
           mj               = Igroup(j);                        %    length mj of vector   Y^j = (H_jr1,... ,H_jrmj) of group j
           MatReta_dj       = MatReta_d(MatIgroup(j,1:mj),:);   %    MatReta_dj(mj,N_d): realizations of Y^j of length mj = Igroup(j)
           cellMatReta_d{j} = MatReta_dj;          
       end  
       clear mj MatReta_argj
                                                                          %--- generation of random Wienner germs for the Ito equation for each group
       cellGauss               = cell(ngroup,1);
       cellWiennerM0transientG = cell(ngroup,1); 
       
       for j = 1:ngroup                                         
           mj                         = Igroup(j);                        %  length mj of vector   Y^j = (H_jr1,... ,H_jrmj) of group j  
           cellGauss{j}               = randn(mj,N_d,nbMC);               %  ArrayGaussj(mj,N_d,nbMC)
           cellWiennerM0transientG{j} = randn(mj,N_d,M0transient,nbMC);   %  ArrayWiennerM0transientj(mj,N_d,M0transient,nbMC)
       end      
       
       disp('--- beginning the construction of parameters for the learning of each group')
       
                                                                          %--- construction of MatReta_ar_g(nu,NnbMC)      
       RepsilonDIFFg  = zeros(ngroup,1);
       RmDPg          = zeros(ngroup,1);
       RmDMAPg        = zeros(ngroup,1);                                  % maximum value of mDMAP for group j
       RnbmDMAPg      = zeros(ngroup,1);                                  % number of DMAP vector retained for group j
       cellIndg       = cell(ngroup,1);                                   % indices of the DMAP vectors retained for group j
       Rsg            = zeros(ngroup,1);
       Rshssg         = zeros(ngroup,1);
       Rshg           = zeros(ngroup,1);
       RDeltarg       = zeros(ngroup,1);
       RM0estimg      = zeros(ngroup,1);
       Rximeang       = zeros(ngroup,1);
       Rf0            = zeros(ngroup,1);
       
      
       parfor j = 1:ngroup                                               %--- loop on the groups
           mj = Igroup(j);                                               %    length mj of vector   Y^j = (H_jr1,... ,H_jrmj) of group j
           MatReta_dj = cellMatReta_d{j};                                %    MatReta_dj(mj,N_d): realizations of Y^j of length mj = Igroup(j)
                                                                         %    MatReta_argj(mj,NnbMC)

           [epsilonDIFFj,mDPj,mDMAPj,nbmDMAPj,Indgj,sj,shssj,shj,Deltarj,M0estimj,ximeanj,f0j] = ...
                                         sub_diffusion_maps_manifold_groupj_parameters(nu,N_d,j,mj,MatReta_dj,ind_generator,icorrectif, ...
                                         Facteur,f0_ref,ind_f0,ind_DMAPS,ind_bandwidth_group,epsilonDIFF,mDP,mDMAP,nbmDMAP,Indg, ...
                                         epsilonDIFFmin_groups); 

           RepsilonDIFFg(j) = epsilonDIFFj;
           RmDPg(j)         = mDPj;
           RmDMAPg(j)       = mDMAPj;                           % maximum value of mDMAP for group j                       
           RnbmDMAPg(j)     = nbmDMAPj;                         % number of DMAP vector retained for group j
           cellIndg{j}      = Indgj;                            % indices of the DMAP vectors retained for group j 
           Rsg(j)           = sj;
           Rshssg(j)        = shssj;
           Rshg(j)          = shj;
           RDeltarg(j)      = Deltarj;
           RM0estimg(j)     = M0estimj;
           Rximeang(j)      = ximeanj;
           Rf0(j)           = f0j;
       end
      
      disp('--- end of the construction of parameters for the learning of each group')   
      
      disp('--- beginning the learning for each group')
      
      cellMatReta_ar = cell(ngroup,1);                          %    cellMatReta_ar{j} =  MatReta_argj(mj,NnbMC)
      cellRlambda    = cell(ngroup,1);                          %    cellRlambda{j}    =  Rlambda(mDPj,1)         
      Rd2mopt        = zeros(ngroup,1);                         %    Rd2mopt(ngroup,1); 
      
      for j = 1:ngroup                                          %--- loop on the groups (DO NOT PARALLELIZE THIS LOOP)
       
          mj           = Igroup(j);                             %    length mj of vector   Y^j = (H_jr1,... ,H_jrmj) of group j
          MatReta_dj   = cellMatReta_d{j};                      %    MatReta_dj(mj,N_d): realizations of Y^j of length mj = Igroup(j)
          epsilonDIFFj = RepsilonDIFFg(j);
          mDPj         = RmDPg(j);
          mDMAPj       = RmDMAPg(j);                            % maximum value of mDMAP for group j  
          nbmDMAPj     = RnbmDMAPg(j);                          % number of DMAP vector retained for group j
          Indgj        = cellIndg{j};                           % indices of the DMAP vectors retained for group j 
          sj           = Rsg(j);
          shssj        = Rshssg(j);
          shj          = Rshg(j);
          Deltarj      = RDeltarg(j);
          M0estimj     = RM0estimg(j);
          ximeanj      = Rximeang(j);
          f0j          = Rf0(j);
          
          ArrayGaussj              = cellGauss{j};               %  ArrayGaussj(mj,N_d,nbMC)
          ArrayWiennerM0transientj = cellWiennerM0transientG{j}; %  ArrayWiennerM0transientj(mj,N_d,M0transient,nbMC)
         
          if ind_constraints == 0                                %--- the constraints E{H] = 0 and E{H H'} = [I_nu] is not applied
                                                                 %    MatReta_ar(nu,NnbMC)
                                                                 %    MatRx_ar(n_d,NnbMC)    
                                                                 %    MatReta_argj(mj,NnbMC)                                                                  
             [MatReta_argj,RLambdagj] = sub_diffusion_maps_manifold_groupj(N_d,j,mj,MatReta_dj,epsilonDIFFj,tdiffusion,mDPj,mDMAPj, ...
                                        nbmDMAPj,Indgj,f0j,M0transient,nbMC,NnbMC,ind_generator,shssj,shj,Deltarj, ...
                                        ArrayWiennerM0transientj,ArrayGaussj,ind_DMAPS,nbmDMAP,MatRg,Rlambda);  
          end
          
          if ind_constraints >= 1                                %--- constraints are applied 
                                                                 %    MatReta_ar(nu,NnbMC)
                                                                 %    MatRx_ar(n_d,NnbMC)    
                                                                 %    MatReta_argj(mj,NnbMC)                                                                  
             [MatReta_argj,RLambdagj] = sub_diffusion_maps_manifold_groupj_c(N_d,j,mj,MatReta_dj,epsilonDIFFj,tdiffusion,mDPj,mDMAPj, ...
                                        nbmDMAPj,Indgj,f0j,M0transient,nbMC,NnbMC,ind_generator,shssj,shj,Deltarj, ...
                                        ArrayWiennerM0transientj,ArrayGaussj,ind_constraints,ind_coupling,epsc,iter_limit, ...
                                        Ralpha_relax_step5,minVarH,maxVarH,ind_DMAPS,nbmDMAP,MatRg,Rlambda);  
          end
          
          cellMatReta_ar{j} =  MatReta_argj; 
          cellRlambda{j}    =  RLambdagj;
                                                         %--- Estimation of the measure concentration computing the L2 norm d2mopt_argj 
          d2mopt_argj = 0; 
          ArrayH_argj  = reshape(MatReta_argj,mj,N_d,nbMC);                                    % ArrayH_argj(mj,N_d,nbMC), MatReta_argj(mj,NnbMC)
          for ell = 1:nbMC
              d2mopt_argj = d2mopt_argj + (norm(ArrayH_argj(:,:,ell) - MatReta_dj,'fro'))^2;   %   MatReta_dj(mj,N_d)
          end
          clear ArrayH_argj
          deno2j      = (norm(MatReta_dj,'fro'))^2;
          d2mopt_argj = d2mopt_argj/(nbMC*deno2j);
          Rd2mopt(j)  = d2mopt_argj;  
          
          disp('------  d2mopt_argj = '); disp(d2mopt_argj) 
      end  
      clear mj MatReta_dj MatReta_argj cellMatReta_d
      clear ArrayGaussj ArrayWiennerM0transientj cellGauss cellWiennerM0transientG 
      
      disp('--- end of the learning for each group')
                                                     %--- plot Rd2mopt(ngroup,1)
      h = figure; 
      hold on
      plot((1:1:ngroup)',Rd2mopt,'b-o','LineWidth',1)
      title('$j\mapsto d^2_{\rm{ar}}(m_{\rm{opt}} ; j)$','FontWeight','Normal','FontSize',16,'Interpreter','latex');
      xlabel('group number j','FontSize',16,'Interpreter','latex')              
      ylabel('$d^2_{\rm{ar}}(m_{\rm{opt}} ; j)$','FontSize',16,'Interpreter','latex')
      saveas(h,'figure_STEP5_d2mopt.fig');   
      hold off
      close(h);
                                                    
      
      if ind_DMAPS == 1                              %--- print: a part of the print have already been printed before  
         fidlisting=fopen('listing.txt','a+');                     
         for j = 1:ngroup
            mDPj         = RmDPg(j);
            Indgj        = cellIndg{j};
            Rnumgj       = (1:1:mDPj)'; 
            RLambdagj    = cellRlambda{j};     
            d2mopt_argj  = Rd2mopt(j);
            
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n ');                     
            fprintf(fidlisting,'------ parameters for group j =  %6i --------------------------------------------------------\n ', j);   
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' mj            = %7i \n ',Igroup(j));
            fprintf(fidlisting,' epsilonDIFFj  = %14.7e \n ',RepsilonDIFFg(j)); 
            fprintf(fidlisting,' mDPj          = %7i \n ',RmDPg(j));
            fprintf(fidlisting,' mDMAPj        = %7i \n ',RmDMAPg(j));          % maximum value of mDMAP for group j  
            fprintf(fidlisting,' nbmDMAPj      = %7i \n ',RnbmDMAPg(j));        % number of DMAP vector retained for group j 
            fprintf(fidlisting,'      \n ');  
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' sj       = %14.7e \n ',Rsg(j));                    
            fprintf(fidlisting,' shssj    = %14.7e \n ',Rshssg(j)); 
            fprintf(fidlisting,' shj      = %14.7e \n ',Rshg(j)); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' f0j      = %14.7e \n ',Rf0(j)); 
            fprintf(fidlisting,' Deltarj  = %14.7e \n ',RDeltarg(j)); 
            fprintf(fidlisting,' M0estimj = %7i \n ',RM0estimg(j));       
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' xi_meanj = %14.7e \n ',Rximeang(j)); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' L2 norm related to the concentration of the measure \n ');
            fprintf(fidlisting,'             \n '); 
            fprintf(fidlisting,'        d^2(m_opt)_argj  =  %14.7e \n',d2mopt_argj);
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n '); 
         end
         fclose(fidlisting);
      end
      
      if ind_DMAPS == 2                              %--- print
         fidlisting=fopen('listing.txt','a+');                     
         for j = 1:ngroup
            mDPj        = RmDPg(j);
            Indgj       = cellIndg{j};
            Rnumgj      = (1:1:mDPj)'; 
            RLambdagj   = cellRlambda{j};  
            d2mopt_argj = Rd2mopt(j);

            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n ');                     
            fprintf(fidlisting,'------ parameters for group j =  %6i --------------------------------------------------------\n ', j);   
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' mj            = %7i \n ',Igroup(j));
            fprintf(fidlisting,' epsilonDIFFj  = %14.7e \n ',RepsilonDIFFg(j)); 
            fprintf(fidlisting,' mDPj          = %7i \n ',RmDPg(j));
            fprintf(fidlisting,' mDMAPj        = %7i \n ',RmDMAPg(j));          % maximum value of mDMAP for group j  
            fprintf(fidlisting,' nbmDMAPj      = %7i \n ',RnbmDMAPg(j));        % number of DMAP vector retained for group j 
            fprintf(fidlisting,'      \n ');           
            fprintf(fidlisting,' Indgj =          \n ');                        % indices of the DMAP vectors retained for group j
            fprintf(fidlisting,' %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i %3i \n ',Indgj');
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' sj       = %14.7e \n ',Rsg(j));                    
            fprintf(fidlisting,' shssj    = %14.7e \n ',Rshssg(j)); 
            fprintf(fidlisting,' shj      = %14.7e \n ',Rshg(j)); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' f0j      = %14.7e \n ',Rf0(j)); 
            fprintf(fidlisting,' Deltarj  = %14.7e \n ',RDeltarg(j)); 
            fprintf(fidlisting,' M0estimj = %7i \n ',RM0estimg(j));       
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' xi_meanj = %14.7e \n ',Rximeang(j)); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n ');  
            fprintf(fidlisting,' Eigenvalues of the transition matrix \n ');                        
            fprintf(fidlisting,'               \n '); 
            MatRprint = [Rnumgj RLambdagj]';                                        
            fprintf(fidlisting,'%5i %18.11e  \n',MatRprint);  
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,' L2 norm related to the concentration of the measure \n ');
            fprintf(fidlisting,'             \n '); 
            fprintf(fidlisting,'        d^2(m_opt)_argj  =  %14.7e \n',d2mopt_argj); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'      \n '); 
         end
         fclose(fidlisting);
      end
      
      MatReta_tilde_arg = zeros(nu,NnbMC);                    %---concatenation of  MatReta_tilde_argj(mj,NnbMC) into MatReta_tilde_arg(nu,NnbMC) 
      for j = 1:ngroup 
          MatReta_tilde_argj = cellMatReta_ar{j};             % MatReta_tilde_argj(mj,NnbMC)
          mj = Igroup(j);                                     % length mj of vector   Y^j = (H_jr1,... ,H_jrmj) of group j
          MatReta_tilde_arg(MatIgroup(j,1:mj),:) = MatReta_tilde_argj; % MatReta_tilde_argj(mj,NnbMC),MatIgroup(ngroup,nu)
      end      
      clear MatReta_tilde_argj cellMatReta_ar
      
      if irenormalization == 0                                %--- No re-normalization 
         MatReta_arg  = MatReta_tilde_arg;                    %    MatReta_arg(nu,NnbMC)
         clear MatReta_tilde_arg
      end
      if irenormalization == 1                                %--- re-normalization of ArrayH_tilde_ar(nu,N_d,nbMC) in ArrayH_ar(nu,N_d,nbMC)  
         Rmean_tilde    = mean(MatReta_tilde_arg,2);          %    Rmean_tilde(nu,1),MatReta_tilde_arg(nu,NnbMC) 
         MatRcov_tilde  = cov(MatReta_tilde_arg');            %    MatRcov_tilde(nu,nu)
         MatRL_tilde    = chol(MatRcov_tilde);                %    MatRcov_tilde(nu,nu) = MatRL_tilde'*MatRL_tilde
         MatRinv        = inv(MatRL_tilde');                  %    MatRinv(nu,nu) 
         MatRmean_tilde = repmat(Rmean_tilde,1,NnbMC);        %    Rmean_tilde(nu,1),MatRmean_tilde(nu,NnbMC)
         MatReta_arg    = MatRinv*(MatReta_tilde_arg - MatRmean_tilde); %    MatReta_tilde_arg(nu,NnbMC),MatReta_arg(nu,NnbMC), MatRmean_tilde(nu,NnbMC)
         clear  MatReta_tilde_arg Rmean_tilde  MatRcov_tilde  MatRL_tilde MatRinv MatRmean_tilde  
      end
      if irenormalization == 2                                %--- Partial re-normalization: only centering and diagonal normalization to 1 of MatReta_ar(nu,NnbMC) 
         Rmean_tilde = mean(MatReta_tilde_arg,2);             %    Rmean_tilde(nu,1),MatReta_tilde_arg(nu,NnbMC) 
         Rstd_tilde  = std(MatReta_tilde_arg,0,2);            %    Rstd_tilde(nu,1)
         MatReta_arg = zeros(nu,NnbMC);                       %    MatReta_arg(nu,NnbMC) 
         for k = 1:nu
             MatReta_arg(k,:) = (MatReta_tilde_arg(k,:) - repmat(Rmean_tilde(k,:),1,NnbMC))./Rstd_tilde(k);  %    MatReta_arg(nu,NnbMC) 
         end
         clear  MatReta_tilde_arg Rmean_tilde Rstd_tilde  
      end   
                                                         %--- Estimation of the measure concentration computing the L2 norm 
                                                         %    d2mopt_arg = E{|| [H_arg] - [eta_d] ||^2} / || [eta_d] ||^2
      d2mopt_arg  = 0; 
      ArrayH_arg  = reshape(MatReta_arg,nu,N_d,nbMC);                                %    ArrayH_arg(nu,N_d,nbMC), MatReta_arg(nu,NnbMC)
      for ell = 1:nbMC
          d2mopt_arg = d2mopt_arg + (norm(ArrayH_arg(:,:,ell) - MatReta_d,'fro'))^2; %    MatReta_d(nu,N_d)
      end
      clear ArrayH_arg
      deno2      = (norm(MatReta_d,'fro'))^2;
      d2mopt_arg = d2mopt_arg/(nbMC*deno2);             
      
      disp('------  d2mopt_arg = '); disp(d2mopt_arg)
      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'-------   L2 norm related to the concentration of the measure \n ');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'          Estimation of E{[H_ar]}: d^2(m_opt)_arg  =  %14.7e \n',d2mopt_arg); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);  
      
      ind_print_arg = 1;
      if ind_print_arg == 1                                                    %--- the print is bypassed if ind_print_arg = 0 
         RHmean_arg    = mean(MatReta_arg,2);                                  %    RHmean_arg(nu,1)
         MatRHcov_arg  = cov(MatReta_arg');                                    %    MatRHcov_arg(nu,nu)  
         fidlisting=fopen('listing.txt','a+');  
         fprintf(fidlisting,'MEAN VECTOR AND COVARIANCE MATRICES OF  H_arg  \n ');                        
         fprintf(fidlisting,'               \n ');
         fprintf(fidlisting,'               \n ');  
         fprintf(fidlisting,'----- RHmean_arg =          \n '); 
         fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',RHmean_arg');
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'----- diag(MatRHcov_arg) =          \n '); 
         fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e ... \n ',diag(MatRHcov_arg));
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'----- MatRHcov_arg =          \n '); 
         for k=1:nu       
             fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',MatRHcov_arg(k,:));
             fprintf(fidlisting,'      \n ');
         end   
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n ');       
         fclose(fidlisting); 
         
         clear RHmean_arg MatRHcov_arg  
      end                                                                 
      
      %-----------------------------------------------------------------------------------------------------------------------------------------                           
      %      plot the addditional realizations of random vectors  MatReta_arg(nu,NnbMC) and  MatRxx_arg(n_d,NnbMC) generated by the PLoM 
      %      using the independent groups
      %----------------------------------------------------------------------------------------------------------------------------------------- 

      sub_plot_diffusion_maps_manifold_H_arg(N_d,NnbMC,MatReta_d,MatReta_arg,nbplotHgSamples,RplotHgSamples,nbplotHg,RplotHg, ...
                                             nbplotHgClouds,MatRplotHgClouds);  
      
      if nbplotXX > 0 || nbplotXXClouds > 0                               %--- construction of MtRetaxx_arg(n_d,NnbMC)
         MatRXmean = repmat(RXmean,1,NnbMC);                              %  MatRXmean(n_d,NnbMC) 
         MatRx_arg = MatRXmean + MatRphiXmu1s2*MatReta_arg;               %  MatRXmean(n_d,NnbMC), MatRphiXmu1s2(n_d,nu),MatReta_arg(nu,NnbMC) 
                                                                          %---  Back scaling:  MatRxx_ar(n_d,NnbMC)
         MatRxx_arg = zeros(n_d,NnbMC);
         for k =1:n_d                                           
             MatRxx_arg(k,:) = Ralpha(k)*MatRx_arg(k,:) + Rbeta(k);   
         end 
         clear MatRXmean MatRx_arg 
         sub_plot_diffusion_maps_manifold_XX_arg(N_d,NnbMC,MatRxx_d,MatRxx_arg,nbplotXX,RplotXX,nbplotXXClouds,MatRplotXXClouds);
      end
      
      ElapsedTime5 = toc(TimeStart5);      
      fidlisting=fopen('listing.txt','a+');
      fprintf(fidlisting,'      \n ');                                                                
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'-------   ELAPSED TIME FOR STEP 5\n ');
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'Elapsed Time   =  %10.2f\n',ElapsedTime5);   
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);                        
      
      SAVERAND = rng;
      
      if isave == 5                                 %--- SAVE after STEP 5 ..............................
         close all                                  %
         save  matlabSTEP5_PLoM_Indep_Group.mat -v7.3
      end                                           %
   end                                              %--- End of ``if irestore < 5 ''
end                                                 %--- End of ``if iexec_Step5 == 1 ''  
close all