
function [INDEPGaussmax,meanINDEPGauss,stdINDEPGauss,gust,meanMaxINDEPGauss,stdMaxINDEPGauss,RINDEPGauss,numfig] = ...
                                                            sub_constructing_GAUSS_reference(NKL,nr,MatRHexpGauss,numfig)
%
% Copyright C. Soize 19 January 2016 
%
%
%---INPUT variables
%
%   nr = number of independent realizations of random vector H
%   MatRHexpGauss  = matrice (NKL,nr) des nr ralisations indpendantes du vecteur alatoire HGauss = (HGauss_1,...,HGauss_NKL)
%   r1  et r2  = indices pour valuer le critre d'indpendance statistique de HGauss_r1 avec HGauss_r2
%
%---OUTPUT variables
%
%   INDEPGauss         : random variable whose realizations are INDEPr1r2Gauss with respect to 1 <= r1 < r2 <= NKL 
%   INDEPGaussmax      : maximum observer on all the realizations of INDEPGauss 
%   MaxINDEPGauss      : random variable corresponding to the exterme values of INDEPGauss
%   meanINDEPGauss     : mean value of  INDEPGauss
%   stdINDEPGauss      : std of  INDEPGauss
%   gust               : gust factor
%   meanMaxINDEPGauss  : mean value of MaxINDEPGauss 
%   stdMaxINDEPGauss   : std of MaxINDEPGauss 
%   RINDEPGauss        : vector of the NKL*(NKL-1)/2 values INDEPr1r2Gauss with respect to 1 <= r1 < r2 <= NKL 
%                        If p = (r1,r2), RINDEPGauss(p) = INDEPr1r2Gauss 
%
%    (1) La solution Gauss indpendante est construite pour avoir la rfrence numrique du
%        critere d'indpendance avec le mme nombre de ralisations : nr    
%    (2) pour chaque couple icouple=(r1,r2), INDEPr1r2Gauss = Sr1Gauss + Sr2Gauss -  Sr1r2Gauss
%        avec Sr1Gauss l'entropie de HGauss_r1, Sr2Gauss l'entropie de HGauss_r2 et                                                           
%             Sr1r2Gauss l'entropie conjointe de HGauss_r1 et HGauss_r2. 
%
%--- METHOD used
%
%    Numerical criterion for testing the independence of two normalized Gaussian random variables HGauss_r1 and HGauss_r2 by using 
%    the MUTUAL INFORMATION criterion: 
%
%    INDEPr1r2Gauss = Sr1Gauss + Sr2Gauss - Sr1r2Gauss > = 0
%    
%    in which Sr1Gauss, Sr2Gauss, and Sr1r2Gauss are the entropy of  HGauss_r1, HGauss_r2, and (HGauss_r1,HGauss_r2).
%    As the Gaussian random vector (HGauss_r1,HGauss_r2) is normalized, we should have INDEPr1r2Gauss  = 0. However as the entropy are 
%    estimated with the Monte Carlo method, we have INDEPr1r2Gauss > 0 (and not = 0).
%    This positive value could thus used as a numerical reference for testing the independence of the non-Gaussian normalized random variables H_r1 and H_r2
%    However, as the number of realization is relatively small, the estmation of the entropy by the Monte Carlo method is not converges and 
%    there are statistical fluctuations. Consequently, INDEPGaussRef is constructed as 
%  
%    Let Z be the positive valued random variable for which the np = NKL*(NKL-1)/2 realizations are z_p = RINDEPGauss(p) = INDEPr1r2Gauss 
%    with 1 <= r1 < r2 <= NKL.  

   np = NKL*(NKL-1)/2;
   RINDEPGauss = zeros(np,1);              % RINDEPGauss(np,1)
   Indr1 = zeros(np,1);
   Indr2 = zeros(np,1);
   
   p = 0;
   for r1 = 1:NKL-1
       for r2 = r1+1:NKL 
           p = p + 1;
           Indr1(p) = r1;
           Indr2(p) = r2;
       end
   end 
  
   parfor p = 1:np
       r1 = Indr1(p);
       r2 = Indr2(p);
       [INDEPr1r2Gauss] = sub_testINDEPr1r2GAUSS(NKL,nr,MatRHexpGauss,r1,r2); 
       RINDEPGauss(p) = INDEPr1r2Gauss;
   end  
   
   meanINDEPGauss = mean(RINDEPGauss);                    % empirical mean value of Z
   stdINDEPGauss  = std(RINDEPGauss);                     % empirical standard deviation Z
   INDEPGaussmax  = max(RINDEPGauss);                     % maximum on the realizations z_p for p=1,...,np
   RINDEPGauss0   = RINDEPGauss - meanINDEPGauss;         % centering 
   nup = 0;
   for p = 2:np                                           % number of upcrossings by 0
       if RINDEPGauss0(p)- RINDEPGauss0(p-1) > 0 && RINDEPGauss0(p)*RINDEPGauss0(p-1) < 0    % there is one upcrossing by zero
          nup = nup + 1;
       end
   end
   cons = sqrt(2*log(nup));
   gust = cons + 0.577/cons;
   meanMaxINDEPGauss = meanINDEPGauss + gust*stdINDEPGauss;  
   stdMaxINDEPGauss  = stdINDEPGauss*(pi/sqrt(6))/cons;    
   
   h = figure;                                            %--- plot the trajectory
   plot(RINDEPGauss,'-b')
         title({['Graph of i^\nu(G_{r_1},G_{r_2}) as a function of the pair (r_1,r_2)']},'FontSize',16,'Interpreter','tex','FontWeight','normal')                                        
         xlabel(['number of the pair (r_1,r_2)'],'FontSize',16,'Interpreter','tex')                                                                     
         ylabel(['i^\nu(G_{r_1},G_{r_2})'],'FontSize',16,'Interpreter','tex')                                                           
         numfig = numfig+1;                                                         
         saveas(h,'figure_STEP4_GROUP_i-Gauss.fig')
    close(h);                                                     
                                                          %--- plot the pdf in assuming the independence of earch pair (r1,r2)
    h = figure; 
    npoint=1000;
    [Rpdf,Rp]  = ksdensity(RINDEPGauss,'npoints',npoint,'support','positive');     
           plot(Rp,Rpdf,'-b')
           title({['Graph of the pdf of Z^\nu whose realizations are '] ; ['z^\nu_p = i^\nu(G_{r_1},G_{r_2}) with p=(r_1,r_2) (blue solid)']},'FontSize',16,'Interpreter','tex','FontWeight','normal')                                        
           xlabel(['z'],'FontSize',16,'Interpreter','tex')                                                                
           ylabel(['pdf_{Z^\nu}(z)'],'FontSize',16,'Interpreter','tex') 
           saveas(h,'figure_STEP4_GROUG_pdf_i-Gauss.fig') 
    close(h);
return                                                                        
