
 function  [MatReta_ar]   = sub_diffusion_maps_manifold_for_parallel(nu,N_d,MatReta_d,epsilonDIFF,tdiffusion,mDP,mDMAP,nbmDMAP,Indg,f0, ...
                                                  M0transient,nbMC,NnbMC,ind_generator,shss,sh,Deltar,ArrayWiennerM0transient,ArrayGauss,irenormalization) 
%                                             
%-----------------------------------------------------------------------------------------------------------------------------------------
% Copyright C. Soize, 10 October 2020
%------------------------------------------------------------------------------------------------------------------------------------------
%
% Publications:  C. Soize, R. Ghanem, Data-driven probability concentration and sampling on manifold, 
%                         Journal of Computational Physics,  doi:10.1016/j.jcp.2016.05.044, 321, 242-258 (2016).
%                C. Soize, Optimal partition in terms of independent random vectors of any non-Gaussian vector defined by a set of realizations, 
%                         SIAM-ASA Journal on Uncertainty Quantification}, doi: 10.1137/16M1062223, 5(1), 176-211 (2017).
%                C. Soize, R. Ghanem, Probabilistic learning on manifolds, Foundations of Data Science, 
%                          American  Institute of Mathematical Sciences (AIMS),  doi: 10.3934/fods.2020013, online,  1-29 (2020). 
%--------------------------------------------------------------------------------------------------------------------------------------------
%
%--- INPUT  
%          N_d           : number of realizations in the initial dataset
%          MatReta_d     : MatReta_d(nu,N_d): realizations of Y^j of length mj
%          epsilonDIFF   :  eps for constructing the kernel in the diffusion maps
%          tdiffusion    :  tdiffusion = 1 (recommended value).
%          mDP           :  Maximum number of dominant diffusion-map vectors computed (number of dominant eigenvalues)
%          mDMAP         :  Dimension of the projection such that mDMAP <= mDP
%          nbmDMAP       :  number of DMAP vectors retained: nbmDMAP <= mDMAP <= mDP
%          Indg          :  Indg(nbmDMAP ,1) contains the indices of the diffusion-map vectors used for the projection of the ISDE    
%          f0            :  Damping coefficient for killing the transient solution 
%          M0transient   :  Number of Deltar introduced before the acquisition of the first samping point (recommendend value: M0transient = 100
%          nbMC          :  NnbMC = N_d*nbMC is the number of additional realizations (for instance nbMC = 100) including the initial dataset                        
%          NnbMC         :  NnbMC = N_d*nbMC number of total additional realizations of random vector H that is generated
%                           including the initial dataset. If N_d is large, do not choose a large value of nbMC.
%          ind_generator:      = 0 generator without using the projection of the ISDE on the diffusion maps basis
%                              = 1 generator using the projection of the ISDE on the diffusion maps basis 
%          shss,sh,Deltar
%          ArrayWiennerM0transient(nu,N_d,M0transient,nbMC)
%          ArrayGauss(nu,N_d,nbMC)
%          irenormalization   = 0 : no normalization
%                             = 1 : renormalzation  (centering and covariance equal to identity)
%                             = 2 : partial renormalization (centering)
%--- OUTPUT
%          MatReta_ar(nu,NnbMC)
%          MatRg(N_d,nbmDMAP)
%          MatRa(N_d,nbmDMAP)
%
%----------------------------------------------------------------------------------------------------------------------------------- 
                                                                   
      
      %------------------------------------------------------------------------------------------------------------------------------------------  
      %                   Stage 1:  computing the diffusion maps  MatRg(N_d,nbmDMAP) associated with MatReta_d(nu,N_d)
      %------------------------------------------------------------------------------------------------------------------------------------------ 

      if ind_generator == 0                                         %--- no projection on the diffusion maps basis
         mDMAP   = N_d;
         nbmDMAP = N_d;
         MatRg   = eye(N_d,N_d);
         Rlambda = ones(N_d,1);
      end
      
      if ind_generator == 1                                          %--- projection on the diffusion maps basis
        [MatRKernel] = sub_Kernel1(N_d,nu,MatReta_d,epsilonDIFF);
                                                                     %    removing the negative zeros of the eigenvalues of MatRKernel
         scaling = max(max(MatRKernel));
         MatRKernel = MatRKernel + scaling*1e-12*eye(N_d,N_d);
                                                                     %    compute the transition probability matrix MatRP and its symmetrized matrix MatRPs
         MatRPs = zeros(N_d,N_d);                                    %    MatRPs(N_d,N_d)
         Rd   = sum(MatRKernel,2);
         Rdm1s2 = 1./sqrt(Rd);
         for i=1:N_d
             MatRPs(i,:) = Rdm1s2(i)*MatRKernel(i,:).*Rdm1s2';       %    MatRPs(N_d,N_d), Symmetric matrix
         end
         clear MatRKernel Rd
         MatRPs = 0.5*(MatRPs+MatRPs');                              %    MatRPs(N_d,N_d), imposing a perfect symmetry
                                           
                                                                     %    compute MatRphi : MatRPs * MatRphi = MatRphi * diag(Rlambda) 

         if 2*mDP+8 >= N_d                                           % computation with eig
            [MatRphi1,MatRdiag] =  eig(MatRPs);                      % MatRPs(N_d,N_d)
            Rlambda1   = diag(MatRdiag);                             % Rlambda1(N_d,1)
            [~,Index]  = sort(1./abs(Rlambda1));                     % modulus in descending order
            Rlambda    = Rlambda1(Index);                            % Rlambda(N_d,1)                          
            MatRphi    = MatRphi1(:,Index);                          % MatRphi(N_d,N_d)
            if mDP + 1 <= N_d
               Rlambda(mDP+1:N_d)   = [];                            % Rlambda(mDP,1)                          
               MatRphi(:,mDP+1:N_d) = [];                            % MatRphi(N_d,mDP)
            end
         end

         if 2*mDP+8 < N_d                                           % computation with eigs
            OPTS.tol   = 1e-10;                                     % convergence tolerance
            OPTS.p     = 2*mDP+8;                                   % dimension of the Arnoldi basis
            OPTS.maxit = 300;                                       % maximum number of iterations
            OPTS.disp  = 0;                                         % number of eigenvalues displayed at each iteration
            [MatRphi1,MatRdiag] =  eigs(MatRPs,mDP,'LM',OPTS);      % MatRphi1(N_d,mDP)
            Rlambda1 = diag(MatRdiag);                              % Rlambda1(mDP,1)
           [~,Index]  = sort(1./abs(Rlambda1));                     % modulus in descending order
            Rlambda = Rlambda1(Index);                              % Rlambda(mDP,1)                          
            MatRphi = MatRphi1(:,Index);                            % MatRphi(N_d,mDP)
         end
         clear Rlambda1 Rbid Index  MatRPhi1 MatRdiag
         clear MatRPs 
         
         fidlisting=fopen('listing.txt','a+');  
         fprintf(fidlisting,'                                              \n '); 
         fprintf(fidlisting,'EIGENVALUES OF THE TRANSITION MATRIX  \n ');                        
         fprintf(fidlisting,'               \n '); 
         Rnum=(1:1:mDP)'; 
         MatRprint = [Rnum Rlambda(1:1:mDP)]';                                        
         fprintf(fidlisting,'%5i %18.11e  \n',MatRprint);  
         fclose(fidlisting); 
                                                                    % Plot
         hold off 
         h=figure;                                
         semilogy((1:1:mDP)',Rlambda(1:1:mDP),'-o') 
         title('Eigenvalues $\Lambda_\alpha$ of the transition matrix in $\log_{10}$ scale','FontSize',16,'Interpreter','latex')                                         
         xlabel('$\alpha$','FontSize',16,'Interpreter','latex')                                                                
         ylabel('$\Lambda_\alpha$','FontSize',16,'Interpreter','latex')
         saveas(h,'figure_STEP3_eigenvalueDMAPS.fig'); 
         hold off
         close(h)
                                                                    % compute MatRPsi : MatRP * MatRPsi = MatRPsi * diag(Rlambda)
         MatRPsi = zeros(N_d,mDP);                            
         for k=1:mDP
             MatRPsi(:,k) = MatRphi(:,k).*Rdm1s2;        
         end  
         Rlambdat = Rlambda.^tdiffusion; 
         clear MatRphi Rdm1s2
                                                                    % loading vectors of the diffusion maps
         MatRg = zeros(N_d,nbmDMAP);                               % MatRg(N_d,nbmDMAPj) selecting the nbmDMAPj diffusion maps with Indgj(1),...,Indgj(nbmDMAPj)      
         for beta=1:nbmDMAP                                       
             MatRg(:,beta) = Rlambdat(Indg(beta))*MatRPsi(:,Indg(beta));              % MatRg(N_d,nbmDMAP) 
         end
         clear Rlambdat MatRPsi                             
      end                            %--- end if indgenerator = 1
                                     %--- generation  of the diffusion-maps basis and checking the projection with the covariance matrix
   
      MatRa = MatRg/(MatRg'*MatRg);  % MatRa(N_d,nbmDMAP): [a] =[g]([g]'[g])^(-1) 
        
      %------------------------------------------------------------------------------------------------------------------------------------------  
      %                   Stage 2:  generation  of NnbMC realizations  MatReta_ar(nu,NnbMC) loaded in ArrayH_ar(nu,N_d,nbMC)  
      %------------------------------------------------------------------------------------------------------------------------------------------                                               

                                             %--- solving the Ito stochastic differential equation in MatRZ
                                             %    ArrayH_ar(nu,N_d,nbMC) = [MatReta_ar(nu,N_d)^1 ... MatReta_ar(nu,N_d)^nbMC)] 
                                                             
                                             %    ArrayZ_ar(nu,nbmDMAP,nbMC)
                                             %    ArrayWiennerM0transient(nu,N_d,M0transient,nbMC),ArrayGauss(nu,N_d,nbMC) 
                                             
      ArrayZ_ar = zeros(nu,nbmDMAP,nbMC);          %    ArrayZ_ar(nu,nbmDMAP,nbMC)
      
      parfor ell = 1:nbMC 
         MatRGauss_ell               = ArrayGauss(:,:,ell);                   %  ArrayGauss(nu,N_d,nbMC)
         ArrayWiennerM0transient_ell = ArrayWiennerM0transient(:,:,:,ell);    %  ArrayWiennerM0transient(nu,N_d,M0transient,nbMC)
         
         [MatRZ_ar] = sub_integration_Verlet_MC_for_parallel(ell,nu,nbmDMAP,N_d,M0transient,Deltar,f0,MatReta_d,MatRa,MatRg,shss,sh, ...
                                                              ArrayWiennerM0transient_ell,MatRGauss_ell);                                                    
                                                             
          ArrayZ_ar(:,:,ell) = MatRZ_ar;                                   
      end
      clear MatRZ_ar MatRGauss_ell ArrayWiennerM0transient_ell ArrayGauss ArrayWiennerM0transient
                                                                 %--- computing ArrayH_tilde_ar(nu,N_d,nbMC)
      ArrayH_tilde_ar = zeros(nu,N_d,nbMC);                                           
      for ell=1:nbMC
             ArrayH_tilde_ar(:,:,ell) = ArrayZ_ar(:,:,ell)*MatRg';     %  ArrayH_tilde_ar(nu,N_d,nbMC), ArrayZ_ar(nu,nbmDMAP,nbMC), MatRg(N_d,nbmDMAP) 
      end
      
      if irenormalization == 0                                   %--- No re-normalization and loading ArrayH_tilde_ar(nu,N_d,nbMC) into ArrayH_ar(nu,N_d,nbMC)
         ArrayH_ar   =  ArrayH_tilde_ar; 
         MatReta_ar  = reshape(ArrayH_ar,nu,NnbMC);               %    MatReta_ar(nu,NnbMC)
         clear ArrayH_tilde_ar ArrayH_ar
      end
      if irenormalization == 1                                   %--- re-normalization of ArrayH_tilde_ar(nu,N_d,nbMC) in ArrayH_ar(nu,N_d,nbMC)                                     
         MatReta_tilde_ar = reshape(ArrayH_tilde_ar,nu,NnbMC);   %    MatReta_tilde_ar(nu,NnbMC) 
         Rmean_tilde = mean(MatReta_tilde_ar,2);                 %    Rmean_tilde(nu,1)
         MatRcov_tilde = cov(MatReta_tilde_ar');                 %    MatRcov_tilde(nu,nu)
         MatRL_tilde   = chol(MatRcov_tilde);                    %    MatRcov_tilde(nu,nu) = MatRL_tilde'*MatRL_tilde
         MatRinv = inv(MatRL_tilde');                            %    MatRinv(nu,nu) 
         ArrayH_ar = zeros(nu,N_d,nbMC);                         %    ArrayH_ar(nu,N_d,nbMC); 
         MatRmean_tilde = repmat(Rmean_tilde,1,N_d);             %    MatRmean_tilde(nu,N_d)
         for ell=1:nbMC
             ArrayH_ar(:,:,ell) = MatRinv*(ArrayH_tilde_ar(:,:,ell) - MatRmean_tilde);
         end
         MatReta_ar    = reshape(ArrayH_ar,nu,NnbMC);               %    MatReta_ar(nu,NnbMC) 
         clear  ArrayH_tilde_ar ArrayH_ar MatReta_tilde_ar Rmean_tilde  MatRcov_tilde  MatRL_tilde MatRinv MatRmean_tilde  
      end
      if irenormalization == 2                                   %--- Partial re-normalization: only centering and diagonal normalization to 1 of MatReta_ar(nu,NnbMC)  
         MatReta_tilde_ar = reshape(ArrayH_tilde_ar,nu,NnbMC);   %    MatReta_tilde_ar(nu,NnbMC) 
         Rmean_tilde = mean(MatReta_tilde_ar,2);                 %    Rmean_tilde(nu,1)
         Rstd_tilde  = std(MatReta_tilde_ar,0,2);                %    Rstd_tilde(nu,1)
         MatReta_ar  = zeros(nu,NnbMC);                          %    MatReta_ar(nu,NnbMC) 
         for k = 1:nu
             MatReta_ar(k,:) = (MatReta_tilde_ar(k,:) - repmat(Rmean_tilde(k,:),1,NnbMC))./Rstd_tilde(k);  %    MatReta_ar(nu,NnbMC) 
         end
         clear  ArrayH_tilde_ar MatReta_tilde_ar Rmean_tilde MatRmean_tilde  
      end                                                      
      
      ind_print_ar = 1;
      if ind_print_ar == 1                                        %--- The print of the mean and the covariance matrix of H_ar are bypassed if ind_print_ar = 0 
          RHmean_ar     = mean(MatReta_ar,2);
          MatRHcov_ar   = cov(MatReta_ar'); 
          fidlisting=fopen('listing.txt','a+'); 
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'      \n ');
          fprintf(fidlisting,'----- RHmean_ar =          \n '); 
          fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',RHmean_ar');
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'----- diag(MatRHcov_ar) =          \n '); 
          fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e ... \n ',diag(MatRHcov_ar));
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'----- MatRHcov_ar =          \n '); 
          for k=1:nu       
              fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',MatRHcov_ar(k,:));
              fprintf(fidlisting,'      \n ');
          end   
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'      \n '); 
          fclose(fidlisting); 
          clear RHmean_ar MatRHcov_ar 
      end  
      return
 end
   


 

 