
 function  [MatReta_ar]   = sub_diffusion_maps_manifold_for_parallel_c(nu,N_d,MatReta_d,epsilonDIFF,tdiffusion,mDP,mDMAP,nbmDMAP,Indg,f0, ...
                                                  M0transient,nbMC,NnbMC,ind_generator,shss,sh,Deltar,ArrayWiennerM0transient,ArrayGauss,irenormalization, ...
                                                  ind_constraints,ind_coupling,epsc,iter_limit,Ralpha_relax,minVarH,maxVarH) 
%                                             
%-----------------------------------------------------------------------------------------------------------------------------------------
% Copyright C. Soize, 08 November 2020
%------------------------------------------------------------------------------------------------------------------------------------------
%
% Publications:  C. Soize, R. Ghanem, Data-driven probability concentration and sampling on manifold, 
%                         Journal of Computational Physics,  doi:10.1016/j.jcp.2016.05.044, 321, 242-258 (2016).
%                C. Soize, Optimal partition in terms of independent random vectors of any non-Gaussian vector defined by a set of realizations, 
%                         SIAM-ASA Journal on Uncertainty Quantification}, doi: 10.1137/16M1062223, 5(1), 176-211 (2017).
%                C. Soize, R. Ghanem, Probabilistic learning on manifolds, Foundations of Data Science, 
%                          American  Institute of Mathematical Sciences (AIMS),  doi: 10.3934/fods.2020013, online,  1-29 (2020). 
%                C. Soize, R. Ghanem, Physics-constrained non-Gaussian probabilistic learning on manifolds, 
%                          International Journal for Numerical Methods in Engineering, doi: 10.1002/nme.6202, 121 (1), 110-145 (2020).
%--------------------------------------------------------------------------------------------------------------------------------------------
%
%--- INPUT  
%          N_d           : number of realizations in the initial dataset
%          MatReta_d     : MatReta_d(nu,N_d): realizations of Y^j of length mj
%          epsilonDIFF   :  eps for constructing the kernel in the diffusion maps
%          tdiffusion    :  tdiffusion = 1 (recommended value).
%          mDP           :  Maximum number of dominant diffusion-map vectors computed (number of dominant eigenvalues)
%          mDMAP         :  Dimension of the projection such that mDMAP <= mDP
%          nbmDMAP       :  number of DMAP vectors retained: nbmDMAP <= mDMAP <= mDP
%          Indg          :  Indg(nbmDMAP ,1) contains the indices of the diffusion-map vectors used for the projection of the ISDE    
%          f0            :  Damping coefficient for killing the transient solution 
%          M0transient   :  Number of Deltar introduced before the acquisition of the first samping point (recommendend value: M0transient = 100
%          nbMC          :  NnbMC = N_d*nbMC is the number of additional realizations (for instance nbMC = 100) including the initial dataset                        
%          NnbMC         :  NnbMC = N_d*nbMC number of total additional realizations of random vector H that is generated
%                           including the initial dataset. If N_d is large, do not choose a large value of nbMC.
%          ind_generator:      = 0 generator without using the projection of the ISDE on the diffusion maps basis
%                              = 1 generator using the projection of the ISDE on the diffusion maps basis 
%          shss,sh,Deltar
%          ArrayWiennerM0transient(nu,N_d,M0transient,nbMC)
%          ArrayGauss(nu,N_d,nbMC)
%          irenormalization  = 0 : no re-normalization after learning (recommended if ind_constraints = 1)
%                            = 1 : re-normalization after learning in order that E{H_ar]=0 and E{H_ar H_ar'}=[I_nu] (recommended if ind_constraints = 0
%                            = 2 : re-normalization after learning in order that E{H_ar]=0 
%                                  ---- Parameters for the convergence of the iteration algorithm for ind_constraints >= 1
%          ind_constraints  = 0 : no constraints E{H] = 0 and E{H H'} = [I_nu]
%                           = 1 : constraints E{H_j^2} = 1 for j =1,...,nu   (highly recommended)
%                           = 2 : constraints E{H] = 0 and E{H_j^2} = 1 for j =1,...,nu
%                           = 3 : constraints E{H] = 0 and E{H H'} = [I_nu]
%          ind_coupling     = 0 : for ind_constraints = 2 or 3, no coupling between the Lagrange multipliers in matrix MatRGammaS_iter 
%                           = 1 : for ind_constraints = 2 or 3, coupling,  all the extra-diagonal blocs in matrix MatRGammaS_iter are kept
%          epsc             = relative tolerance for the iteration convergence in the constraints E{H} = 0 and E{H H'} = [I_nu] 
%          iter_limit       = maximum number of iteration for computing the Lagrange multiplier
%          Ralpha_relax     = Ralpha_relax(iter_limit,1) relaxation parameter in ] 0 , 1 ] for the iteration. 
%          minVarH          = minimum imposed on Var{H^2} with respect to 1 (for instance 0.99) for the convergence test
%          maxVarH          = maximum imposed on Var{H^2} with respect to 1 (for instance 1.01) for the convergence test
%--- OUTPUT
%          MatReta_ar(nu,NnbMC)
%
%----------------------------------------------------------------------------------------------------------------------------------- 
                                                                   
      
      %------------------------------------------------------------------------------------------------------------------------------------------  
      %                   Stage 1:  computing the diffusion maps  MatRg(N_d,nbmDMAP) associated with MatReta_d(nu,N_d)
      %------------------------------------------------------------------------------------------------------------------------------------------ 

      if ind_generator == 0                                         %--- no projection on the diffusion maps basis
         mDMAP   = N_d;
         nbmDMAP = N_d;
         MatRg   = eye(N_d,N_d);
         Rlambda = ones(N_d,1);
      end
      
      if ind_generator == 1                                          %--- projection on the diffusion maps basis
        [MatRKernel] = sub_Kernel1(N_d,nu,MatReta_d,epsilonDIFF);
                                                                     %    removing the negative zeros of the eigenvalues of MatRKernel
         scaling = max(max(MatRKernel));
         MatRKernel = MatRKernel + scaling*1e-12*eye(N_d,N_d);
                                                                     %    compute the transition probability matrix MatRP and its symmetrized matrix MatRPs
         MatRPs = zeros(N_d,N_d);                                    %    MatRPs(N_d,N_d)
         Rd   = sum(MatRKernel,2);
         Rdm1s2 = 1./sqrt(Rd);
         for i=1:N_d
             MatRPs(i,:) = Rdm1s2(i)*MatRKernel(i,:).*Rdm1s2';       %    MatRPs(N_d,N_d), Symmetric matrix
         end
         clear MatRKernel Rd
         MatRPs = 0.5*(MatRPs+MatRPs');                              %    MatRPs(N_d,N_d), imposing a perfect symmetry
                                           
                                                                     %    compute MatRphi : MatRPs * MatRphi = MatRphi * diag(Rlambda) 

         if 2*mDP+8 >= N_d                                           % computation with eig
            [MatRphi1,MatRdiag] =  eig(MatRPs);                      % MatRPs(N_d,N_d)
            Rlambda1   = diag(MatRdiag);                             % Rlambda1(N_d,1)
            [~,Index]  = sort(1./abs(Rlambda1));                     % modulus in descending order
            Rlambda    = Rlambda1(Index);                            % Rlambda(N_d,1)                          
            MatRphi    = MatRphi1(:,Index);                          % MatRphi(N_d,N_d)
            if mDP + 1 <= N_d
               Rlambda(mDP+1:N_d)   = [];                            % Rlambda(mDP,1)                          
               MatRphi(:,mDP+1:N_d) = [];                            % MatRphi(N_d,mDP)
            end
         end

         if 2*mDP+8 < N_d                                           % computation with eigs
            OPTS.tol   = 1e-10;                                     % convergence tolerance
            OPTS.p     = 2*mDP+8;                                   % dimension of the Arnoldi basis
            OPTS.maxit = 300;                                       % maximum number of iterations
            OPTS.disp  = 0;                                         % number of eigenvalues displayed at each iteration
            [MatRphi1,MatRdiag] =  eigs(MatRPs,mDP,'LM',OPTS);      % MatRphi1(N_d,mDP)
            Rlambda1 = diag(MatRdiag);                              % Rlambda1(mDP,1)
           [~,Index]  = sort(1./abs(Rlambda1));                     % modulus in descending order
            Rlambda = Rlambda1(Index);                              % Rlambda(mDP,1)                          
            MatRphi = MatRphi1(:,Index);                            % MatRphi(N_d,mDP)
         end
         clear Rlambda1 Rbid Index  MatRPhi1 MatRdiag
         clear MatRPs 
         
         fidlisting=fopen('listing.txt','a+');  
         fprintf(fidlisting,'                                              \n '); 
         fprintf(fidlisting,'EIGENVALUES OF THE TRANSITION MATRIX  \n ');                        
         fprintf(fidlisting,'               \n '); 
         Rnum=(1:1:mDP)'; 
         MatRprint = [Rnum Rlambda(1:1:mDP)]';                                        
         fprintf(fidlisting,'%5i %18.11e  \n',MatRprint);  
         fclose(fidlisting); 
         
         hold off 
         h=figure;                                
         semilogy((1:1:mDP)',Rlambda(1:1:mDP),'-o') 
         title('Eigenvalues $\Lambda_\alpha$ of the transition matrix in $\log_{10}$ scale','FontSize',16,'Interpreter','latex')                                         
         xlabel('$\alpha$','FontSize',16,'Interpreter','latex')                                                                
         ylabel('$\Lambda_\alpha$','FontSize',16,'Interpreter','latex')
         saveas(h,'figure_STEP3_eigenvalueDMAPS.fig'); 
         hold off
         close(h)
                                                                             % compute MatRPsi : MatRP * MatRPsi = MatRPsi * diag(Rlambda)
         MatRPsi = zeros(N_d,mDP);                            
         for k=1:mDP
             MatRPsi(:,k) = MatRphi(:,k).*Rdm1s2;        
         end  
         Rlambdat = Rlambda.^tdiffusion; 
         clear MatRphi Rdm1s2
                                                                   % loading vectors of the diffusion maps
         MatRg = zeros(N_d,nbmDMAP);                               % MatRg(N_d,nbmDMAPj) selecting the nbmDMAPj diffusion maps with Indgj(1),...,Indgj(nbmDMAPj)      
         for beta=1:nbmDMAP                                       
             MatRg(:,beta) = Rlambdat(Indg(beta))*MatRPsi(:,Indg(beta));              % MatRg(N_d,nbmDMAP) 
         end
         clear Rlambdat MatRPsi                             
      end                            %--- end if indgenerator = 1
                                     %--- generation  of the diffusion-maps basis and checking the projection with the covariance matrix
   
      MatRa = MatRg/(MatRg'*MatRg);  % MatRa(N_d,nbmDMAP): [a] =[g]([g]'[g])^(-1) 
        
      %------------------------------------------------------------------------------------------------------------------------------------------  
      %                   Stage 2:  generation  of NnbMC realizations  MatReta_ar(nu,NnbMC) loaded in ArrayH_ar(nu,N_d,nbMC) 
      %                             with the constraints
      %------------------------------------------------------------------------------------------------------------------------------------------ 
                                         
      if ind_constraints == 1                   %--- constraints E{H_j^2} = 1 for j =1,...,nu  
         mhc = nu;
         Rbc = ones(nu,1);                      %    Rbc(mhc,1) :  E{h^c(H)} = b^c corresponding to  E{H_j^2} = 1 for j =1,...,nu
         Rlambda_iter_m1 = zeros(mhc,1);        %    Rlambda_iter_m1(mhc,1) = ( lambda_1,...,lambda_nu )
         MatRlambda_iter_m1 =[]; 
      end
      if ind_constraints == 2                   %--- constraints E{H} = 0 and E{H_j^2} = 1 for j =1,...,nu
         mhc = 2*nu;
         Rbc = zeros(mhc,1);                    %    Rbc(mhc,1) :  E{h^c(H)} = b^c corresponding to E{H} = 0 and E{H_j^2} = 1 for j =1,...,nu
         Rbc(1+nu:2*nu,1) = ones(nu,1); 
         Rlambda_iter_m1 = zeros(mhc,1);        %    Rlambda_iter_m1(mhc,1) = (lambda_1,...,lambda_nu,lambda_{1+nu},...,lambda_{nu+nu})
         MatRlambda_iter_m1 =[];      
      end
      if ind_constraints == 3                   %--- constraints E{H] = 0 and E{H H'} = [I_nu]
         mhc = 2*nu + nu*(nu-1)/2;              %    note that for nu = 1, we have mhc = 2*nu
                                                %    if nu = 1: Rbc(mhc,1) = (bc_1,...,bc_nu , bc_{1+nu},...,bc_{nu+nu} )
                                                %    if nu > 1: Rbc(mhc,1) = (bc_1,...,bc_nu , bc_{1+nu},...,bc_{nu+nu} , {bc_{(j,i) + 2 nu} for 1 <= j < i <= nu} )
         Rbc = zeros(mhc,1);                    %    Rbc(mhc,1) :  E{h^c(H)} = b^c corresponding to E{H] = 0  and  E{H H'} = [I_nu] 
         Rbc(1+nu:2*nu,1) = ones(nu,1); 
         Rlambda_iter_m1  = zeros(mhc,1);       %    Rlambda_iter_m1(mhc,1)      
         if nu == 1                             %    if nu = 1: Rlambda_iter(mhc,1) = (lambda_1,...,lambda_nu,lambda_{1+nu},...,lambda_{nu+nu})
            MatRlambda_iter_m1 =[];
         end    
         if nu >=2                              %    if nu > 1: Rlambda_iter(mhc,1) = ( lambda_1,...,lambda_nu , lambda_{1+nu},...,lambda_{nu+nu} ,
                                                %                                      {lambda_{(j,i) + 2 nu} for 1 <= j < i <= nu} )
            MatRlambda_iter_m1 = zeros(nu,nu);  %    MatRlambda_iter_m1(nu,nu) = Symmetric with zero diagonal and upper part is  Rlambda_iter  row rise
            ind = 0;
            for j = 1:nu-1
                for i = j+1:nu
                    ind = ind + 1;
                    MatRlambda_iter_m1(j,i) = Rlambda_iter_m1(2*nu+ind);
                    MatRlambda_iter_m1(i,j) = MatRlambda_iter_m1(j,i);
                end
            end
        end
      end
      
      fidlisting=fopen('listing.txt','a+'); 
      fprintf(fidlisting,'      \n '); 
      fprintf(fidlisting,'mhc           = %7i \n ',mhc); 
      fprintf(fidlisting,'      \n '); 
      fclose(fidlisting);  
          
      %--- Computation of the Lagrange multipliers by the iteration algorithm  and generation  of nbMC realizations 
                                                          
      Rerr               = zeros(iter_limit,1);   %    Rerr(iter_limit,1)      
      RnormRlambda       = zeros(iter_limit,1);   %    RnormRlambda(iter_limit,1)
      RcondGammaS        = zeros(iter_limit,1);   %    RcondGammaS(iter_limit,1)
      Rtol               = zeros(iter_limit,1);   %    Rtol(iter_limit,1) 
      Rtol(1)            = 1;
      ind_conv           = 0;
         
      for iter = 1:iter_limit                              %    LOOP of the iteration algorithm (iter : lambda_iter given and compute lambda_iterplus1
                                                           %---------------------------------------------------------------------------------------------
          disp('-------------iter'); disp(iter)            %    For lambda = lambda_iter, solving the Ito stochastic differential equation in MatRZ
                                                           %    ArrayH_ar(nu,N_d,nbMC) = [MatReta_ar(nu,N_d)^1 ... MatReta_ar(nu,N_d)^nbMC)] 
                                                           %    with  MatReta_ar(nu,N_d)^ell the realization for ell = 1,...,nbMC   
                                                           %--- Computing additional realizations 
                                                           %    MatReta_ar_iter(nu,nbMC) 
          
                                                           %--- solving the Ito stochastic differential equation in  ArrayZ_ar_iter(nu,nbmDMAP,nbMC)
                                                           %  
          if ind_constraints == 1                          %--- constraints E{H_j^2} = 1 for j =1,...,nu  
             ArrayZ_ar_iter = zeros(nu,nbmDMAP,nbMC);      %    ArrayZ_ar_iter(nu,nbmDMAP,nbMC)
             parfor ell = 1:nbMC 
                MatRGauss_ell               = ArrayGauss(:,:,ell);                   %  ArrayGauss(nu,N_d,nbMC)
                ArrayWiennerM0transient_ell = ArrayWiennerM0transient(:,:,:,ell);    %  ArrayWiennerM0transient(nu,N_d,M0transient,nbMC)

               [MatRZ_ar_iter] = sub_integration_Verlet_MC_for_parallel_c1(ell,nu,nbmDMAP,N_d,M0transient,Deltar,f0,MatReta_d,MatRa,MatRg,shss,sh, ...
                                                                          ArrayWiennerM0transient_ell,MatRGauss_ell,mhc,Rlambda_iter_m1);                                                    

                ArrayZ_ar_iter(:,:,ell) = MatRZ_ar_iter;  %   ArrayZ_ar_iter(nu,nbmDMAP,nbMC)                                
             end
             clear MatRZ_ar_iter MatRGauss_ell ArrayWiennerM0transient_ell 
          end
          
          if ind_constraints == 2                          %--- constraints E{H_j} = 0 for j =1,...,nu and E{H_j^2} = 1 for j =1,...,nu
             ArrayZ_ar_iter = zeros(nu,nbmDMAP,nbMC);      %    ArrayZ_ar_iter(nu,nbmDMAP,nbMC)
             parfor ell = 1:nbMC 
                MatRGauss_ell               = ArrayGauss(:,:,ell);                   %  ArrayGauss(nu,N_d,nbMC)
                ArrayWiennerM0transient_ell = ArrayWiennerM0transient(:,:,:,ell);    %  ArrayWiennerM0transient(nu,N_d,M0transient,nbMC)

               [MatRZ_ar_iter] = sub_integration_Verlet_MC_for_parallel_c2(ell,nu,nbmDMAP,N_d,M0transient,Deltar,f0,MatReta_d,MatRa,MatRg,shss,sh, ...
                                                                          ArrayWiennerM0transient_ell,MatRGauss_ell,mhc,Rlambda_iter_m1);                                                    

                ArrayZ_ar_iter(:,:,ell) = MatRZ_ar_iter;  %   ArrayZ_ar_iter(nu,nbmDMAP,nbMC)                                
             end
             clear MatRZ_ar_iter MatRGauss_ell ArrayWiennerM0transient_ell 
          end
          
          if ind_constraints == 3                          %--- constraints E{H] = 0 and E{H H'} = [I_nu]  
             ArrayZ_ar_iter = zeros(nu,nbmDMAP,nbMC);      %    ArrayZ_ar_iter(nu,nbmDMAP,nbMC)
             parfor ell = 1:nbMC 
                MatRGauss_ell               = ArrayGauss(:,:,ell);                   %  ArrayGauss(nu,N_d,nbMC)
                ArrayWiennerM0transient_ell = ArrayWiennerM0transient(:,:,:,ell);    %  ArrayWiennerM0transient(nu,N_d,M0transient,nbMC)

               [MatRZ_ar_iter] = sub_integration_Verlet_MC_for_parallel_c3(ell,nu,nbmDMAP,N_d,M0transient,Deltar,f0,MatReta_d,MatRa,MatRg,shss,sh, ...
                                                                          ArrayWiennerM0transient_ell,MatRGauss_ell,mhc,Rlambda_iter_m1,MatRlambda_iter_m1);                                                    

                ArrayZ_ar_iter(:,:,ell) = MatRZ_ar_iter;  %   ArrayZ_ar_iter(nu,nbmDMAP,nbMC)                                
             end
             clear MatRZ_ar_iter MatRGauss_ell ArrayWiennerM0transient_ell 
          end
          
          ArrayH_ar_iter = zeros(nu,N_d,nbMC);                       %--- computing ArrayH_ar_iter(nu,N_d,nbMC)                                         
          for ell=1:nbMC
                 ArrayH_ar_iter(:,:,ell) = ArrayZ_ar_iter(:,:,ell)*MatRg';  %  ArrayH_ar_iter(nu,N_d,nbMC), ArrayZ_ar_iter(nu,nbmDMAP,nbMC), MatRg(N_d,nbmDMAP) 
          end
                                                                          %--- computing h^c(H) loaded in MatRhc_iter(mhc,NnbMC)
          MatReta_ar_iter          = reshape(ArrayH_ar_iter,nu,NnbMC);    %    MatReta_ar_iter(nu,NnbMC)
          clear ArrayH_ar_iter   
          
          testNaN =isnan(norm(MatReta_ar_iter,'fro')); 
          if testNaN  >= 1                                                %   there is NaN , indetermination of type 0/0 in the ISDE solver          
             disp('----- STOP: indetermination of type NaN or Inf in the ISDE solver ')
             disp('iter'); disp(iter);
             disp('Rlambda_iter_m1');disp(Rlambda_iter_m1');
             fidlisting=fopen('listing.txt','a+');
             fprintf(fidlisting,'      \n ');
             fprintf(fidlisting,'      \n ');
             fprintf(fidlisting,'----- STOP: indetermination of type of type NaN or Inf in the ISDE solver \n '); 
             fprintf(fidlisting,'      \n '); 
             fprintf(fidlisting,'      \n ');             
             fprintf(fidlisting,'      iter             = %7i \n ',iter);
             fprintf(fidlisting,'      Rlambda_iter_m1  = %14.7e %14.7e %14.7e %14.7e %14.7e %14.7e %14.7e %14.7e \n ',Rlambda_iter_m1');
             fprintf(fidlisting,'      \n ');
             fprintf(fidlisting,'      \n ');
             fprintf(fidlisting,'      Decrease alpha_relax2 or increase iter_relax2\n ');
             fprintf(fidlisting,'      \n ');
             fprintf(fidlisting,'      \n ');
             fclose(fidlisting); 
             error('STOP')                                                   %     STOP: divergence of ISDE
          end
          
          if ind_constraints == 1                                            %--- constraints E{H_j^2} = 1 for j =1,...,nu
             MatRhc_iter              = zeros(mhc,NnbMC);                    %    MatRhc_iter(mhc,NnbMC);
             MatRhc_iter(1:nu,:)      = MatReta_ar_iter.^2;                  %    MatReta_ar_iter(nu,NnbMC),MatRhc_iter(mhc,NnbMC)
          end
          if ind_constraints == 2                                            %--- constraints E{H} = 0 and E{H_j^2} = 1 for j =1,...,nu
             MatRhc_iter              = zeros(mhc,NnbMC);                    %    MatRhc_iter(mhc,NnbMC);
             MatRhc_iter(1:nu,:)      = MatReta_ar_iter;                     %    MatReta_ar_iter(nu,NnbMC),MatRhc_iter(mhc,NnbMC)
             MatRhc_iter(1+nu:2*nu,:) = MatReta_ar_iter.^2;                  %    MatReta_ar_iter(nu,NnbMC),MatRhc_iter(mhc,NnbMC)
          end
          if ind_constraints == 3                                            %--- constraints E{H] = 0 and E{H H'} = [I_nu]
             MatRhc_iter              = zeros(mhc,NnbMC);                    %    MatRhc_iter(mhc,NnbMC);
             MatRhc_iter(1:nu,:)      = MatReta_ar_iter;                     %    MatReta_ar_iter(nu,NnbMC),MatRhc_iter(mhc,NnbMC)
             MatRhc_iter(1+nu:2*nu,:) = MatReta_ar_iter.^2;                  %    MatReta_ar_iter(nu,NnbMC),MatRhc_iter(mhc,NnbMC)
             if nu >=2  
                ind = 0;
                for j = 1:nu-1
                    for i = j+1:nu
                        ind = ind + 1;
                        MatRhc_iter(2*nu+ind,:) = MatReta_ar_iter(j,:).*MatReta_ar_iter(i,:);
                    end
                end
             end
          end
        
          RVarH_iter          = var(MatReta_ar_iter,0,2);
          minVarH_iter        = min(RVarH_iter );
          maxVarH_iter        = max(RVarH_iter );
          clear RVarH_iter
                                                  
          Rmeanhc_iter          = mean(MatRhc_iter,2);                           %  Rmeanhc_iter(mhc,1),MatRhc_iter(mhc,NnbMC)           
          RGammaP_iter          = Rbc - Rmeanhc_iter;                            %  RGammaP_iter(mhc,1),Rbc(mhc,1),Rmeanhc_iter(mhc,1) 
          MatRGammaS_iter_temp  = cov(MatRhc_iter');                             %  MatRGammaS_iter_temp(mhc,mhc) 
          
          if ind_coupling == 0                                                   %--- For ind_constraints = 2 or 3, no coupling between the Lagrange 
                                                                                 %    multipliers in matrix MatRGammaS_iter 
             if ind_constraints == 1                                             %... constraints E{H_j^2} = 1 for j =1,...,nu
                 MatRGammaS_iter  = MatRGammaS_iter_temp;                        %    MatRGammaS_iter(mhc,mhc)   
             end
             if ind_constraints == 2                                             %... constraints E{H} = 0 and E{H_j^2} = 1 for j =1,...,nu
                MatRGammaS_iter_temp(1:nu,nu+1:nu+nu) = zeros(nu,nu);
                MatRGammaS_iter_temp(nu+1:nu+nu,1:nu) = zeros(nu,nu);
                MatRGammaS_iter  = MatRGammaS_iter_temp;                         %    MatRGammaS_iter(mhc,mhc)  
             end
             if ind_constraints == 3                                             %... constraints E{H] = 0 and E{H H'} = [I_nu]
                MatRGammaS_iter_temp(1:nu,1+nu:mhc) = zeros(nu,mhc-nu);
                MatRGammaS_iter_temp(1+nu:mhc,1:nu) = zeros(mhc-nu,nu);
                MatRGammaS_iter_temp(1+nu:nu+nu,1+2*nu:mhc) = zeros(nu,mhc-2*nu);
                MatRGammaS_iter_temp(1+2*nu:mhc,1+nu:nu+nu) = zeros(mhc-2*nu,nu);
                MatRGammaS_iter  = MatRGammaS_iter_temp;                         %    MatRGammaS_iter(mhc,mhc)  
             end 
             clear MatRGammaS_iter_temp 
          end
          
          Rerr(iter)          = norm(RGammaP_iter)/norm(Rbc);                    %  Rerr(iter_limit,1)  
          RnormRlambda(iter)  = norm(Rlambda_iter_m1);                           %  RnormRlambda(iter_limit,1) 
          RcondGammaS(iter)   = cond(MatRGammaS_iter);                           %  RcondGammaS(iter_limit,1)          
          disp('err_iter ');disp(Rerr(iter));
          disp('norm_lambda_iter ');disp(RnormRlambda(iter));
          disp('condGammaS_iter ');disp(RcondGammaS(iter));
          if iter >= 2
             Rtol(iter) = 2*abs(Rerr(iter)-Rerr(iter-1))/abs(Rerr(iter)+Rerr(iter-1));
             disp('tol_iter '); disp(Rtol(iter));
          end  
          fidlisting=fopen('listing.txt','a+'); 
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'          ---iter number       = %7i \n ',iter);
          fprintf(fidlisting,'             err_iter         = %14.7e \n ',Rerr(iter));
          fprintf(fidlisting,'             tol_iter         = %14.7e \n ',Rtol(iter));
          fprintf(fidlisting,'      \n '); 
          fclose(fidlisting); 
                                                                                     %---convergence is reached, exit from the loop on iter  
          if (iter > 10) && (Rerr(iter)-Rerr(iter-1) > 0)                            %  a local minimum is obtained for iter - 1
             ind_conv   = 1; 
             MatReta_ar = MatReta_ar_iter_m1;                                        %   MatReta_ar(nu,NnbMC), MatReta_ar_iter_m1(nu,NnbMC)
             iter_plot  = iter-1;
             disp('             Convergence with criterion 1')
             fidlisting=fopen('listing.txt','a+'); 
             fprintf(fidlisting,'      \n '); 
             fprintf(fidlisting,'             Convergence with criterion 1 \n ');
             fprintf(fidlisting,'      \n '); 
             fclose(fidlisting); 
             clear MatReta_ar_iter_m1 MatReta_ar_iter
             clear MatRhc_iter Rmeanhc_iter RGammaP_iter Rlambda_iter_m1
             break                          % exit from the loop on iter  
          end
          if ((minVarH_iter >= minVarH) && (maxVarH_iter <= maxVarH)) || Rerr(iter) <= epsc   % the variance of a component H_k is greater than 1 OR the error is less than the tolerance: exit
             ind_conv   = 1; 
             MatReta_ar = MatReta_ar_iter;                                           %   MatReta_ar(nu,NnbMC), MatReta_ar_iter(nu,NnbMC)
             iter_plot  = iter;
             disp('             Convergence with criterion 2')
             fidlisting=fopen('listing.txt','a+'); 
             fprintf(fidlisting,'      \n '); 
             fprintf(fidlisting,'             Convergence with criterion 2 \n ');
             fprintf(fidlisting,'      \n '); 
             fclose(fidlisting); 
             clear MatReta_ar_iter_m1 MatReta_ar_iter
             clear  MatRhc_iter Rmeanhc_iter RGammaP_iter Rlambda_iter_m1
             break                          % exit from the loop on iter
         end                                                                    
         if (iter > min(20,iter_limit) && Rerr(iter) < 0.09 ) &&  (Rtol(iter) < 0.01)     % the error is stationary: exit 
            ind_conv  = 1; 
            MatReta_ar = MatReta_ar_iter;                                       %   MatReta_ar(nu,NnbMC), MatReta_ar_iter(nu,NnbMC)
            iter_plot  = iter;
            disp('             Convergence with criterion 3')
            fidlisting=fopen('listing.txt','a+'); 
            fprintf(fidlisting,'      \n '); 
            fprintf(fidlisting,'             Convergence with criterion 3 \n ');
            fprintf(fidlisting,'      \n '); 
            fclose(fidlisting); 
            clear MatReta_ar_iter_m1 MatReta_ar_iter
            clear MatRhc_iter Rmeanhc_iter RGammaP_iter Rlambda_iter_m1
            break                         % exit from the loop on iter
         end
                                                                                 %---convergence not reached 
         if ((minVarH_iter < minVarH) || (maxVarH_iter > maxVarH)) && Rerr(iter) > epsc
             Rtemp_iter      = MatRGammaS_iter\RGammaP_iter;                     %   Rtemp_iter = inv(MatRGammaS_iter)*RGammaP_iter
             Rlambda_iter    = Rlambda_iter_m1 - Ralpha_relax(iter)*Rtemp_iter;  %   Rlambda_iter(mhc,1), Rlambda_iter_m1(mhc,1)
             MatRlambda_iter =[];
             if ind_constraints == 3 && nu >= 2                                  %--- constraints E{H] = 0 and E{H H'} = [I_nu]
                MatRlambda_iter = zeros(nu,nu);                                  %  MatRlambda_iter(nu,nu) 
                ind = 0;
                for j = 1:nu-1
                    for i = j+1:nu
                        ind = ind + 1;
                        MatRlambda_iter(j,i) = Rlambda_iter(2*nu+ind);
                        MatRlambda_iter(i,j) = MatRlambda_iter(j,i);
                    end
                end
             end             
             Rlambda_iter_m1    =  Rlambda_iter;
             MatRlambda_iter_m1 = MatRlambda_iter;  
             MatReta_ar_iter_m1 = MatReta_ar_iter;
             clear MatRhc_iter Rlambda_iter MatRlambda_iter Rmeanhc_iter RGammaP_iter MatRGammaS_iter  Rtemp_iter
         end 
      end                         %--- end for iter = 1:iter_limit  
      
      if ind_conv == 0 % iter_limit has been reached without convergence
         MatReta_ar = MatReta_ar_iter;                                       %   MatReta_ar(nu,NnbMC), MatReta_ar_iter(nu,NnbMC)
         iter_plot  = iter_limit;       
         disp('------ No convergence of the iteration algorithm')
         disp('       iter_plot = '); disp(iter_plot) 
         fidlisting=fopen('listing.txt','a+'); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n ');
         fprintf(fidlisting,'------ No convergence of the iteration algorithm    \n '); 
         fprintf(fidlisting,'                                                    \n ');
         fprintf(fidlisting,'       iter             = %7i    \n ',iter_plot);
         fprintf(fidlisting,'       err_iter         = %14.7e \n ',Rerr(iter_plot));
         fprintf(fidlisting,'       tol_iter         = %14.7e \n ',Rtol(iter_plot));
         fprintf(fidlisting,'       norm_lambda_iter = %14.7e \n ',RnormRlambda(iter_plot));
         fprintf(fidlisting,'       condGammaS_iter  = %14.7e \n ',RcondGammaS(iter_plot));
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'       Decrease alpha_relax2 or increase iter_relax2\n ');
         fprintf(fidlisting,'      \n '); 
         fclose(fidlisting); 
      end
      if ind_conv == 1 % convergence is reached
         disp('------ Convergence of the iteration algorithm')
         disp('       iter_plot = '); disp(iter_plot)  
         fidlisting=fopen('listing.txt','a+'); 
         fprintf(fidlisting,'      \n '); 
         fprintf(fidlisting,'      \n ');
         fprintf(fidlisting,'------ Convergence of the iteration algorithm    \n '); 
         fprintf(fidlisting,'                                                    \n ');
         fprintf(fidlisting,'       iter             = %7i    \n ',iter_plot);
         fprintf(fidlisting,'       err_iter         = %14.7e \n ',Rerr(iter_plot));
         fprintf(fidlisting,'       tol_iter         = %14.7e \n ',Rtol(iter_plot));
         fprintf(fidlisting,'       norm_lambda_iter = %14.7e \n ',RnormRlambda(iter_plot));
         fprintf(fidlisting,'       condGammaS_iter  = %14.7e \n ',RcondGammaS(iter_plot));         
         fprintf(fidlisting,'      \n '); 
         fclose(fidlisting); 
      end
                                                                %--- plot
      h = figure;      
      plot((1:1:iter_plot)',Rerr(1:iter_plot),'b-')                                                 
      title('Graph of function $\rm{err}(\iota)$','FontSize',16,'Interpreter','latex','FontWeight','normal')                                         
      xlabel('$\iota$','FontSize',16,'Interpreter','latex')                                                                
      ylabel('$\rm{err}(\iota)$','FontSize',16,'Interpreter','latex')     
      saveas(h,'figure_STEP3_Rerr.fig'); 
      close(h);	

      h = figure;      
      plot((1:1:iter_plot)',RnormRlambda(1:iter_plot),'b-')                                                 
      title('Graph of function $\Vert \lambda_{\iota}\Vert $','FontSize',16,'Interpreter','latex','FontWeight','normal')                                         
      xlabel('$\iota$','FontSize',16,'Interpreter','latex')                                                                
      ylabel('$\Vert \lambda_{\iota}\Vert $','FontSize',16,'Interpreter','latex')     
      saveas(h,'figure_STEP3_RnormRlambda.fig'); 
      close(h);	
      
      h = figure;      
      plot((1:1:iter_plot)',RcondGammaS(1:iter_plot),'b-')                                                 
      title('Graph of function $\rm{cond} [\Gamma"(\lambda_{\iota})]$','FontSize',16,'Interpreter','latex','FontWeight','normal')                                         
      xlabel('$\iota$','FontSize',16,'Interpreter','latex')                                                                
      ylabel('$\rm{cond} [\Gamma"(\lambda_{\iota})]$','FontSize',16,'Interpreter','latex')     
      saveas(h,'figure_STEP3_RcondGammaS.fig'); 
      close(h);
      
      MatReta_tilde_ar = MatReta_ar;                                  %    MatReta_tilde_ar(nu,NnbMC), MatReta_ar(nu,NnbMC)
      
      if irenormalization == 0                                         %--- No re-normalization 
         MatReta_ar  = MatReta_tilde_ar;                               %    MatReta_ar(nu,NnbMC)
         clear MatReta_tilde_ar
      end
      if irenormalization == 1                                         %--- re-normalization of ArrayH_tilde_ar(nu,N_d,nbMC) in ArrayH_ar(nu,N_d,nbMC) 
         Rmean_tilde    = mean(MatReta_tilde_ar,2);                    %    Rmean_tilde(nu,1),MatReta_tilde_ar(nu,NnbMC) 
         MatRcov_tilde  = cov(MatReta_tilde_ar');                      %    MatRcov_tilde(nu,nu)
         MatRL_tilde    = chol(MatRcov_tilde);                         %    MatRcov_tilde(nu,nu) = MatRL_tilde'*MatRL_tilde
         MatRinv        = inv(MatRL_tilde');                           %    MatRinv(nu,nu) 
         MatRmean_tilde = repmat(Rmean_tilde,1,NnbMC);                 %    Rmean_tilde(nu,1),MatRmean_tilde(nu,NnbMC)
         MatReta_ar     = MatRinv*(MatReta_tilde_ar - MatRmean_tilde); %    MatReta_tilde_ar(nu,NnbMC),MatReta_ar(nu,NnbMC), MatRmean_tilde(nu,NnbMC)
         clear  MatReta_tilde_ar Rmean_tilde  MatRcov_tilde  MatRL_tilde MatRinv MatRmean_tilde  
      end
      if irenormalization == 2                                   %--- Partial re-normalization: only centering and diagonal normalization to 1 of MatReta_ar(nu,NnbMC)  
         MatReta_tilde_ar = reshape(ArrayH_tilde_ar,nu,NnbMC);   %    MatReta_tilde_ar(nu,NnbMC) 
         Rmean_tilde = mean(MatReta_tilde_ar,2);                 %    Rmean_tilde(nu,1)
         Rstd_tilde  = std(MatReta_tilde_ar,0,2);                %    Rstd_tilde(nu,1)
         MatReta_ar  = zeros(nu,NnbMC);                          %    MatReta_ar(nu,NnbMC) 
         for k = 1:nu
             MatReta_ar(k,:) = (MatReta_tilde_ar(k,:) - repmat(Rmean_tilde(k,:),1,NnbMC))./Rstd_tilde(k);  %    MatReta_ar(nu,NnbMC) 
         end
         clear  ArrayH_tilde_ar MatReta_tilde_ar Rmean_tilde MatRmean_tilde  
      end 
      
      ind_print_ar = 1;
      if ind_print_ar == 1                                        %--- The print of the mean and the covariance matrix of H_ar are bypassed if ind_print_ar = 0 
          RHmean_ar     = mean(MatReta_ar,2);
          MatRHcov_ar   = cov(MatReta_ar'); 
          fidlisting=fopen('listing.txt','a+'); 
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'      \n ');
          fprintf(fidlisting,'----- RHmean_ar =          \n '); 
          fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',RHmean_ar');
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'----- diag(MatRHcov_ar) =          \n '); 
          fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e ... \n ',diag(MatRHcov_ar));
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'----- MatRHcov_ar =          \n '); 
          for k=1:nu       
              fprintf(fidlisting,' %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e \n ',MatRHcov_ar(k,:));
              fprintf(fidlisting,'      \n ');
          end   
          fprintf(fidlisting,'      \n '); 
          fprintf(fidlisting,'      \n '); 
          fclose(fidlisting); 
          clear RHmean_ar MatRHcov_ar 
      end  
      return
 end
   


 

 